/*
 * Copyright (C) 2000-2007 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/*!
 * $Log: sensor-node.c,v $
 * Revision 1.2  2007/01/17 17:00:59  dyerm
 * simplify includes since /extras is in include path
 *
 * Revision 1.1  2007/01/17 15:31:47  dyerm
 * renamed sensor.c -> sensor-node.c
 *
 * Revision 1.2  2007/01/17 15:29:45  dyerm
 * code clean-up
 *
 */

#include "../defs.h"
#include "program_version.h"

// sensor related includes
#include <dev/adc.h>
#include <dev/twif.h>
#include <teco_ssmall/micsampler.h>
#include <teco_ssmall/tsl2550.h>
#include <teco_ssmall/tc74.h>

// address of the base station
bt_addr_t sink_addr = {0x8b,0x01,0x00,0x3f,0x04,0x00};

// sensor thread
THREAD(sensorLoop, arg){  
    // init twi
    TwInit(20); // set our slave addr = 20
    
    // start light sensor
    while(tsl_init()){
        NutSleep(10000);
    }
        
    // init mic
    ADCInit();
    mic_init();
    
    // sensor loop:
    sensor_data_t sensor_data;
    sensor_data_t* data = &sensor_data;
    u_char channel0, channel1;
    signed int temp;
    for(;;){
    	NutSleep(5000);
    	// sample mic
        data->mic = mic_read();
        // sample light
        tsl_read(&channel0, &channel1);
       	data->light = tsl_calculate_lux(channel0, channel1);
		// read temperature
		tc_read(&temp);
		data-> temp = temp;
  		// output
  		printf("mic = %u, light = %d, temp = %u\n",
  				data->mic,
  				data->light,
  				data->temp);
  				
  		// send packet
  		mhop_cl_send_pkt((u_char*)data, sizeof(sensor_data_t), sink_addr, SENSO_PSM, MHOP_CL_UNICAST, MHOP_CL_TTL_INFINITE);
    }  		
}

int main(void)
{
    // hardware init
    btn_hardware_init();
    btn_led_init(1);

    // init terminal app uart
    u_long baud = 57600;     // serial baud rate
    NutRegisterDevice(&APP_UART, 0, 0);
    freopen(APP_UART.dev_name, "r+", stdout);
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    btn_terminal_init(stdout, "[senso]$");

    // hello message
    printf("\n# ------------------------------------------------");
    printf("\n# Welcome to the EWSN 2007 Tutorial (c) ETH Zurich\n");
    printf("# program version: %s\n", PROGRAM_VERSION);
    printf("# ------------------------------------------------");

    printf("\nbooting bluetooth module... ");
    // bluetooth module on (takes a while)
    btn_hardware_bt_on();
    printf("ok.\n\r");
    
    // Start bt-stack and let the initialization begin
    printf("init bt-stack... ");
	struct btstack* bt_stack;
    bt_stack = bt_hci_init(&BT_UART);
    printf("done.\n");

	// Initialize connection-less multi-hop layer
    // Set ACL packet types
    printf("setting acl pkt types... ");
    bt_acl_init(bt_stack, BT_HCI_PACKET_TYPE_DM3);
    printf("done.\n");
    
    // Init protocol/service multiplexor
    printf("init protcol/service mux... ");
	bt_psm_t* psmux;
    psmux = bt_psm_init(bt_stack, MAX_NR_SERVICES, NR_BUFFERS);
    printf("done.\n");
    
    // Init connectionless l2cap stack
    printf("init connectionless l2cap... ");
    l2cap_cl_init(bt_stack, psmux);
    printf("done.\n");
    
    // Init connection manager
    printf("init connection manager... ");
    con_mgr_init(bt_stack, psmux, CM_PSM, bt_hci_register_con_table_cb, CM_COD);
    printf("done.\n");
    
    // Init connectionless multihop protocol
    printf("init connectionless multi-hop protocol... ");
    mhop_cl_init(bt_stack, psmux, MHOP_PSM,	NR_BUFFERS, con_mgr_register_con_table_cb);
    // init rpc
    rpc_init(psmux, 8, RPC_PROC_PSM, RPC_RESULT_PSM);
    mhop_con_info_init();
    printf("done.\n");
	      	
    // sensor board power
    btn_hardware_io_power(1);

	// start sensor thread
	NutThreadCreate("T_senso", sensorLoop, 0, 256);

    // terminal mode
    btn_terminal_run(BTN_TERMINAL_NOFORK, 0);

    return 0;
}
