/*
 * Copyright (C) 2006 Florian Schuetz <fschuetz@ethz.ch>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * For additional information see http://www.btnode.ethz.ch/
 * 
 */

#include <hardware/btn-hardware.h>
#include <dev/usartavr.h>
#include <terminal/btn-terminal.h>
#include <led/btn-led.h>
#include <terminal/btn-cmds.h>
#include <terminal/nut-cmds.h>
#include <terminal/eeprom_cmds.h>
#include <cc/cc1000.h>
#include <cc/bmac.h>
#include <cc/ccc.h>
#include <string.h>
#include <eepromdb/btn-eepromdb.h>
#include <cc/aodv.h>
#include <cc/rfc.h>
#include <terminal/aodv-cmds.h>
#include <terminal/rfc-cmds.h>

#define NODE_ADDRESS 		0x001 //This node address is used when none is stored in EEPROM
#define CCC_RFP                 3     //Power of chipcon (50 wo antenna, 3 with antenna are good values)


/**
 * Returns the configured mac address to the terminal
 *
 * @return configured mac address, 0 if no mac address is stored in EEPROM.
 */
u_short get_MAC_address(void)
{
  u_long address;
  
  if ( btn_eepromdb_get( FOURCC('M','A','C',' '),
			 BTN_EEPROMDB_INT_SIZE,
			 &address ) == 0 )
    {
      return 0;
    }
  
  return (u_short) address;
}
 
 
/**
 * Reads/writes the Mac-Address to the internal EEPROM
 */
void address_cmd(char* arg)
{
  u_char result[100];
  u_long address;
  
  if (strncmp(arg, "set", 3) == 0) {
    if ( sscanf( arg+4, "%s", result ) == 0 ) {
      printf( "error: no parameter!\n" );
      return;                
    }
      
    // convert string to integer
    if ( sscanf( arg+4, "%lx", &address ) <= 0 ) {
      printf( "error: positive number required!\n" );
      return;                                                                
    }
      
    // check if address <= 0xFFFF
    if ( address > 0xFFFF ) {
      printf( "error: address must be between 0x0000 and 0xFFFF!\n" );
      return;                                                                                        
    }
      
    printf( "writing %lu...\n", address );
    
    if ( btn_eepromdb_store( FOURCC('M','A','C',' '), 4, &address) == 0 ) {
      printf( "error while storing!\n" );
      return;
    }
      
    printf( "OK\n" );
      
  } else if (strncmp(arg, "get", 3) == 0) {
    if ( btn_eepromdb_get( FOURCC('M','A','C',' '), 4, &address ) == 0 ) {
      printf( "error!\n" );
      return;
    }
      
    printf( "MAC = 0x%04lx\n", address );
  } else if (strncmp(arg, "del", 3) == 0) {
    if ( btn_eepromdb_delete( FOURCC('M','A','C',' ') ) == 0 ) {
      printf( "error!\n" );
      return;
    }
      
    printf( "OK\n" );
  } else if (strncmp(arg, "auto", 4) == 0) {
    //initialise_MAC_address();
    printf("NOT IMPLEMENTED\n");
  }              
  else {
    printf( "usage: address get\n" );
    printf( "       address set <address>\n" );            
  }
}
 

/**
 * Initializes the terminal
 */
void init_stdout(void) {
	u_long baud = 57600;
	
	NutRegisterDevice(&APP_UART, 0, 0);
	freopen(APP_UART.dev_name, "r+", stdout);
	_ioctl(_fileno(stdout), UART_SETSPEED, &baud);
}

/**
 * Initializes the CC1000 radio.
 */
static void init_radio(u_char rfp) {

  int res;
  u_short address;

  /* Check if there is an address stored in EEPROM, else set default */
  if((address = get_MAC_address()) == 0)
    address = NODE_ADDRESS;


  /* MAC Inititalisation */
  // initialise bmac - also fill bmac_interface structure 
  res = bmac_init(address);
  if(res != 0) {
    //Time to panic - MAC initialisation failed
    printf("BMAC initialisation failed!\n");
  }
  bmac_enable_led(0);

  // Initialise ccc 
  res = ccc_init(&bmac_interface);
  if(res != 0) {
    // Time to panic - cc1000 initialization failed
    printf("CCC Initialisation failed!\n");
  }

  // Set RF-Power
  cc1000_set_RF_power((u_char) rfp);

  /* Routing Initialisation */
  // Initialise bt-aodv
  res = aodv_init(address);
  if(res != 0) {
    //Time to panic - aodv initialization failed
    printf("AODV Initialisation failed!\n");
  }
  aodv_enable_led(1);

  // Intitialze RFC
  res = rfc_init(&aodv_interface);
  if(res != 0) {
    // You know...
    printf("RFC Initialization failed!\n");
  }
  rfc_enable_led(1);
  rfc_enable_ack(1);

  printf("Radio successfully initialised.\n");
}


int main(void) {
  // INSOMNIA! sleep mode causes chipcon reception to collapse
  NutThreadSetSleepMode(SLEEP_MODE_NONE);

  //Initialise 
  btn_hardware_init(); 	/* initialise SRAM */
  btn_led_init(0);	/* initialise LEDs */
  init_stdout();        /* initialise stdout */

  ROUTING_DEBUG_INIT;

  printf("Initializing RFC-TEST...\n");

  //Initialize Radio with receiving power CCC_RFP
  init_radio(CCC_RFP); 
  printf("Node initialisation Complete\n");

  //Initialise terminal
  char prompt[15];
  u_short mac_address = get_MAC_address();
  sprintf(prompt, "[%s@%02X:%02X]$", "rfc-test", mac_address >> 8, 
    	mac_address & 0xFF );
  btn_terminal_init(stdout, prompt);
  
  btn_cmds_register_cmds();
  nut_cmds_register_cmds();
  aodv_cmds_register_cmds();
  rfc_cmds_register_cmds();
  eeprom_cmds_register_cmds();
  btn_terminal_register_cmd("address", address_cmd);

  btn_terminal_run(BTN_TERMINAL_NOFORK, 0);

  return 0;
}
