
/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 * $Log: ram_banks.c,v $
 * Revision 1.12  2008/01/09 19:07:44  freckle
 * off by one was really wrong, end address was correct from the beginning
 *
 * Revision 1.11  2008/01/09 18:48:52  freckle
 * corrected bankcpy_check_addess(address,len) second time (was off by 1
 *
 * Revision 1.10  2008/01/09 17:24:40  freckle
 * corrected bankcpy_check_address function (was using 179 KB as limit) instead of 177 KB
 *
 * Revision 1.9  2007/02/21 14:09:02  freckle
 * Reduced code size and cleaned up
 *
 * Revision 1.8  2006/03/24 15:13:16  freckle
 * added avr-libc version test to choose correct include paths
 *
 * Revision 1.7  2006/03/23 07:22:22  dyerm
 * Merged changes from multihop_merge branch. See individual changes on multihop_merge branch. See Changelog for summary of changes.
 *
 * Revision 1.5.2.1  2006/03/22 14:07:32  dyerm
 * merged changes from main branch from [multihop_merged_1] to [multihop_merged_2] into multihop_merge branch
 *
 * Revision 1.6  2006/03/21 15:40:23  beutel
 * changed includes for avr-libc 1.4.* compatibility
 *
 * Revision 1.5  2005/08/22 13:04:09  lwinterhalter
 * Last revision broke bootloader. Reverted changes.
 * *
 * Revision 1.3  2005/06/03 14:03:43  olereinhardt
 * Re-indented files according to NutOS coding style: -kr -nut -l 132
 *
 * Revision 1.2  2004/10/08 15:13:12  freckle
 * removed .init1 section
 * added crc16 xbank functions
 *
 * Revision 1.1  2004/10/08 14:37:34  freckle
 * Added ram banks functions
 *
 * 2003. 1.10 Urs Frey <ufrey@ee.ethz.ch>
 *
 *
 * functions to write and read from additional RAM banks (addr>64kB)
 *
 */


// -------------------------------------------------------------------------
// Includes
// -------------------------------------------------------------------------

#include <hardware/ram_banks.h>
#include <hardware/btn-hardware.h>
#include <sys/thread.h>
#include <string.h>

#ifdef __AVR_ATmega128__
#if __AVR_LIBC_VERSION__ >= 10400UL
#include <util/crc16.h>
#else
#include <avr/crc16.h>
#endif
#endif

// -------------------------------------------------------------------------
// mem copy routines
//
// address range: 0-183551
// -------------------------------------------------------------------------

#if defined(__AVR_ATmega128__)

extern void *bank_cpy_01(void *, void *, size_t);
extern void *bank_cpy_02(void *, void *, size_t);
extern void *bank_cpy_03(void *, void *, size_t);
extern void *bank_cpy_10(void *, void *, size_t);
extern void *bank_cpy_20(void *, void *, size_t);
extern void *bank_cpy_30(void *, void *, size_t);

#else

void * bank_cpy_01(void *to, void *from, size_t len)
{
    memcpy(bank1 + (u_long) to - 0x1100, from, len);
    return to;
}
void * bank_cpy_02(void *to, void *from, size_t len)
{
    memcpy(bank2 + (u_long) to - 0x1100, from, len);
    return to;
}
void * bank_cpy_03(void *to, void *from, size_t len)
{
    memcpy(bank3 + (u_long) to - 0x1100, from, len);
    return to;
}
void * bank_cpy_10(void *to, void *from, size_t len)
{
    memcpy(to, bank1 + (u_long) from - 0x1100, len);
    return to;
}
void * bank_cpy_20(void *to, void *from, size_t len)
{
    memcpy(to, bank2 + (u_long) from - 0x1100, len);
    return to;
}
void * bank_cpy_30(void *to, void *from, size_t len)
{
    memcpy(to, bank3 + (u_long) from - 0x1100, len);
    return to;
}

#endif



// function table for copying
typedef void * (*bank_copy_function)(void*, void *, size_t);
bank_copy_function   fToXRAM[4] = { NULL, &bank_cpy_01, &bank_cpy_02, &bank_cpy_03 };
bank_copy_function fFromXRAM[4] = { NULL, &bank_cpy_10, &bank_cpy_20, &bank_cpy_30 };



static inline
u_char bankcpy_check_address(u_long address, u_short len)
{
    if (address + len > (XRAM_BANK_SIZE * 3) ) {
        //printf("ERROR: address range from 0x0 to 0x2CCFF"NL);
        return 1;
    }
    return 0;
}

static inline
u_char bankcpy_get_bank(u_long address)
{
    return (u_char) (address / XRAM_BANK_SIZE) + 1;
}

static inline
u_short bankcpy_get_offset(u_long address)
{
    return (u_short) (address % XRAM_BANK_SIZE + 0x1100UL);
}

static inline
u_short bankcpy_get_len_remain(u_long address)
{
    return (u_short) (XRAM_BANK_SIZE - address % XRAM_BANK_SIZE);
}


u_char cpy( void *buffer, u_long address, u_short len, u_char toXRAM) {
    
    bank_copy_function * cpyF;

    u_char start_bank, end_bank;
#if defined(__linux__) || defined (__APPLE__)
    u_long start_offset;
#else
    u_short start_offset;
#endif
    u_short banklen;

    if (bankcpy_check_address(address, len)) {
        return 1;
    }

    if (toXRAM) {
        cpyF = fToXRAM;
    } else {
        cpyF = fFromXRAM;
    }
    
    start_bank = bankcpy_get_bank(address);
    end_bank = bankcpy_get_bank(address + len);
    
    start_offset = bankcpy_get_offset(address);
    
    if (start_bank == end_bank) {
        banklen = len;
    } else if (start_bank == end_bank - 1) {
        banklen = bankcpy_get_len_remain(address);
    } else {
        //printf("ERROR: cpy failed...");
        return 1;
    }
    
    //printf("cpy to bank %u, offset %u, len %u"NL, start_bank, start_offset, banklen);
    (*cpyF[start_bank])( (u_char *) start_offset, buffer, banklen );
    //printf("cpy to bank %u, offset %u, len %u"NL, end_bank, 0x1100, len-banklen);
    if (start_bank == end_bank - 1) {
        (*cpyF[end_bank])( (u_char *) 0x1100, buffer + banklen, len - banklen );
    }
    return 0;
}

u_char cpy_to_xbank(void *buffer, u_long address, u_short len)
{
    return cpy( buffer, address, len, 1);
}

u_char cpy_from_xbank(void *buffer, u_long address, u_short len)
{
    return cpy( buffer, address, len, 0);
}

/* -------------------------------------------------------------------------
*  update crc reverse, needed for unix
* ------------------------------------------------------------------------- */
#ifdef __BTN_UNIX__

#define		M16	0xA001

u_short _crc16_update(u_short crc, u_short c)
{
    u_char i;
    for (i = 0; i < 8; i++) {
        if ((crc ^ c) & 1)
            crc = (crc >> 1) ^ M16;
        else
            crc >>= 1;
        c >>= 1;
    }
    return crc;
}
#endif

/* -------------------------------------------------------------------------
* calculate xbank crc
* ------------------------------------------------------------------------- */
u_short get_xbank_crc(u_long start, u_long end)
{
#define CHECK_BUFFER_LEN 50
    u_char check_buffer[CHECK_BUFFER_LEN];
    u_char i;
    u_char len;
    u_short crc16 = 0;

    //calculate crc
    len = CHECK_BUFFER_LEN;
    while (start < end) {
        if (start + CHECK_BUFFER_LEN > end) {
            len = end - start;
        }
        cpy_from_xbank(check_buffer, start, len);
        for (i = 0; i < len; i++) {
            crc16 = _crc16_update(crc16, check_buffer[i]);
        };
        start += CHECK_BUFFER_LEN;
    }
    
    return crc16;
}

u_char check_xbank_crc(u_long start, u_long end)
{
    return (get_xbank_crc(start, end + CRC_LENGTH) != 0);
}

void update_xbank_crc(u_long start, u_long end)
{

    u_char buffer[2];
    u_short crc = get_xbank_crc(start, end);

    buffer[0] = (u_char) (crc & 0xFF);
    buffer[1] = (u_char) (crc >> 8);

    cpy_to_xbank(buffer, end, 2);

}
