/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch/
 *
 */
 
/**
 * \file bt_hci_api.h
 * \author Mathias Payer     <payerm@stud.ethz.ch>
 * \author Martin Hinz       <btnode@hinz.ch>
 * \author Matthias Ringwald <mringwal@inf.ethz.ch>
 *
 * @brief This headerfile describes the API of the HCI cmd and some get functions to
 * retrieve connection information.
 * 
 * So far only the most important calls are listed.
 */
 
#ifndef _BT_HCI_API_H_
#define _BT_HCI_API_H_

#include <bt/bt_defs.h>
#include <bt/bt_hci_defs.h>

// For now the only possible way is to use the uart driver!
#if HCI_TRANSPORT == UART
	#include <bt/bt_hci_transport_uart.h>
#else
	#include <bt/bt_hci_transport_usb.h>
#endif

/** 
 * for uart debugging. to turn on uart debugging set this value to 1
 */
extern u_char _bt_hci_debug_uart;
/**
 * @struct btstack
 * @brief Represents one entity of a running stack (more than one possible)
 * Keeps all data for one stack, like devices, buffers, states, ...
 * This is the main struct your program should have.
 *
 * To initialize one example stack use:
 *	struct btstack stack;		// The btstack is named stack (e.g.)
 *	struct bt_hci_uart uart;	// The uart for the uart-layer (usb may follow)
 * The above variables should be global!
 *	uart.hwdev = &devUart0;		// Set the uart to the first serial port
 *	stack.transport = &uart;	// And register the uart to the specified stack
 *	bt_hci_init(&stack);		// Start new stack.
 *
 * From now on the btstack is ready and you can register a higher
 * layer (l2cap for example) or just start programming your own
 * lowlevel functions
 * (See bt_hci_l2cap.h for l2cap use and definitions)
 *
 * For possible (callable) HCI-Functions hava a look at the
 * bt_hci_cmds.h file!
 * @file bt_defs.h (there you find the definition for the struct)
*/

/*@{*/

/** \addtogroup Bt_Error_Codes
 *
 * @brief List of HCI error codes. (see bluetooth spec section 6)
 * 
\verbatim

Error Code  Description

0x01 Unknown HCI Command.
0x02 No Connection.
0x03 Hardware Failure.
0x04 Page Timeout.
0x05 Authentication Failure.
0x06 Key Missing.
0x07 Memory Full.
0x08 Connection Timeout.
0x09 Max Number Of Connections.
0x0A Max Number Of SCO Connections To A Device.
0x0B ACL connection already exists.
0x0C Command Disallowed.
0x0D Host Rejected due to limited resources.
0x0E Host Rejected due to security reasons.
0x0F Host Rejected due to remote device is only a personal device.
0x10 Host Timeout.
0x11 Unsupported Feature or Parameter Value.
0x12 Invalid HCI Command Parameters.
0x13 Other End Terminated Connection: User Ended Connection.
0x14 Other End Terminated Connection: Low Resources.
0x15 Other End Terminated Connection: About to Power Off.
0x16 Connection Terminated by Local Host.
0x17 Repeated Attempts.
0x18 Pairing Not Allowed.
0x19 Unknown LMP PDU.
0x1A Unsupported Remote Feature.
0x1B SCO Offset Rejected.
0x1C SCO Interval Rejected.
0x1D SCO Air Mode Rejected.
0x1E Invalid LMP Parameters.
0x1F Unspecified Error.
0x20 Unsupported LMP Parameter Value.
0x21 Role Change Not Allowed
0x22 LMP Response Timeout
0x23 LMP Error Transaction Collision
0x24 LMP PDU Not Allowed
0x25 Encryption Mode Not Acceptable
0x26 Unit Key Used
0x27 QoS is Not Supported
0x28 Instant Passed
0x29 Pairing with Unit Key Not Supported
0x2A-0xFF Reserved for Future Use.

\endverbatim
 * 
 */

#define BT_ERR_UNKNOWN_HCI_COMMAND     -0x01
#define BT_ERR_ROLE_CHANGE_NOT_ALLOWED -0x21

#define BT_HCI_CON_HANDLE_OWN			0

/*
 * @brief Error codes returned by bluetooth api.
 * Errors in the range -0x01 to -0xFF are errors from the bt-controller!
 * See the btspec for the different meanings!
 */
#define BT_ERR_BT_MODULE        -0x0100 /**< problems with bluetooth module                 */
#define BT_ERR_UART_COM         -0x0101	/**< There was an error with the uart-commun.       */
#define BT_ERR_USB_COM          -0x0102	/**< There was an error with the usb-commun.        */
#define BT_ERR_NO_CON           -0x0103 /**< no connection with this app_con_handle exists	*/
#define BT_ERR_MAX_CON          -0x0104 /**< max nr of connection reached                   */
#define BT_ERR_RST_RUN          -0x0105 /**< reset already running                          */
#define BT_ERR_MAX_WAITCMD      -0x0106 /**< stack waitqueue full                           */
#define BT_ERR_CON_EXIST        -0x0107 /**< a connection to bd addr already exists         */
#define BT_ERR_CON_PEND_ACTIVE  -0x0108 /**< a connection to bd addr is actively  opened    */
#define BT_ERR_CON_PEND_PASSIVE -0x0109 /**< a connection to bd addr is passively opened    */
#define BT_ERR_CON_BROKEN       -0x010a /**< disconnected during cmd execution              */
#define BT_ERR_ACL_PLOAD_OVFLW	-0x010b /**< too much data written to an open acl link      */
#define BT_ERR_ACL_WRITE		-0x010c


// ----------------------------------------------------------
// disconnect reasons
// ----------------------------------------------------------
/**
 * Disconnection because of connection link lost
 */
#define BT_HCI_DISCON_LINK_LOST          0x08  //Connection broke

/**
 * Disconnection because other end terminated connection
 */
#define BT_HCI_DISCON_OTHER_DEVICE       0x13  //Other end terminated connection

/**
 * Disconnection because of low ressource
 */
#define BT_HCI_DISCON_LOW_RESSOURCE      0x14  //Device terminates the connection because of low ressource

/**
 * Disconnection because of device shutting down
 */
#define BT_HCI_DISCON_ABOUT_TO_POWER_OFF 0x15  //Device is about to shutdown

/**
 * Disconnection because local device requested
 */
#define BT_HCI_DISCON_BY_LOCAL_DEVICE    0x16  //Local device terminated connection

/**
 * Shortcut to get the maximal payload size of an acl communication packet.
 */
#define BT_ACL_PKT_MAX_PAYLOAD(btstack)		(btstack->host_acl_len)

/**
 *************************************************
 * @struct bt_hci_inquiry_result
 *************************************************
 * This struct defines one entry for a device found during inquiry!
 * The application calls the inquiry cmd with a pointer to an array of these
 * struct entries and the size of this array (max_dev)!
 * All multi-byte values are stored in little endian (over-the-air representation)!
*/
struct bt_hci_inquiry_result {
	bt_addr_t bdaddr;                   /**< The address of the discovered bt-module */
	u_char page_scan_rep_mode :4;       /**< The page-scan repetition mode (0x0, 0x1 or 0x2) */
	u_char page_scan_period_mode :4;	/**< The page-scan period mode (0x0, 0x1 or 0x2) */
	u_long cod;                         /**< The class of device (actually only low 3 bytes -> cod & 0xFFFFFF )*/
	u_short clock_offset;               /**< Clock offset to the other device (15Bit, Bit 16: Valid/Invalid) */
	short rssi;                         /**< rssi, range: -127 to +20, Units: dBm */
};

/**
 *************************************************
 * @struct bt_hci_local_version_result
 *************************************************
 * This struct is used to separate the parameters returned by the local version
*/
struct bt_hci_local_version_result {
	u_char hciversion;
	u_short hcirevision;
	u_char lmpversion;
	u_short manufacturername;
	u_short lmpsubversion;
};


/*
 * array of connection handles, returned by bt_hci_get_con_handles
 */
typedef bt_hci_con_handle_t bt_hci_con_handle_array[BT_HCI_MAX_NUM_CON]; // brauchen wir das wirklich?


/**
 * @struct bt_hci_cmd_response
 *
 * Represents a cmd that is waiting to be completed
 * You only have to mess with this struct if you want some ASYNC calls!
 * If you are OK with plain SYNC stuff, you can disregard it!
 *
 * For asynchronous use of a function call, pass a pointer to this struct. As
 * soon as the event is #SIGNALED, a short response value will be set in the struct.
 * For results of more complex commands (e.g., inquiry), a special struct is used!
 * use NUT-Functions for event in the response struct:
 * - NutEventWait(&x_response->event, time_ms) [time_ms =60-#NUT_WAIT_INFINITE]
 * or
 * - (x_response->event == #SIGNALED)?
 *
 * @see bt_hci_event.c (& h) how the responsed are stored and how the waiting threads are signaled
 * @see the directory bt/hci-cmds for the different commands that can be used with this struct
*/
struct bt_hci_cmd_response {
	u_short ogfocf;		/**< The ogfocf pair (is set by internal functions, for internal use only!) */
	u_short cmd_handle;		/**< The cmd handle (is set by internal functions, for internal use only!) */
	long response;		/**< The response value of the function if max. short, values <0 represent an error! */
	void *ptr;		/**< This is a pointer to a result, if it is bigger than short. */
	HANDLE block;		/**< This is the wait-queue (is set by internal functions, for internal use only!) */
};

/*******************************************************************************
 * bt_setup
 ******************************************************************************/
/**
 * @brief Setups the Bluetooth module
 * 
 * Sets common parameters like baudrate etc.
 *
 */
void bt_setup(struct btstack *stack);

/*******************************************************************************
 * bt_hw_shutdown
 ******************************************************************************/
/**
 * @brief Prepare the BT stack for shutdown and power off the Bluetooth module
 *
 * - lock all semaphores related to BT
 * - unlock all semaphores except command queue
 * - replace the connection request callback with reject
 * - sets the right baudrate, if necessary
 * - power off Bluetooth module with btn_hardware_bt_power_off()
 * - set conn_request and pin_request to 0
 * - emulates disconnect stack-internally (wakeup threads, call con_table callbacks etc.)
 *
 * only available on BTnode3
 *
 * @param stack The bt-stack to use
 * @return pointer to the hci connection request callback (needed by startup)
 */
void * bt_hw_shutdown(struct btstack * stack);

/*******************************************************************************
 * bt_hw_startup
 ******************************************************************************/
/**
 * @brief Doing a hardware reset of the Bluetooth module
 *
 * - power on Bluetooth module with btn_hardware_bt_power_on()
 * - releases semaphore nr_hci_cmds with value 1
 * - call bt_setup
 * - replace the connection request callback with the previous one
 * - unlock command queue
 *
 * only available on BTnode3
 *
 * @param stack The bt-stack to use
 * @param con_rqst_cb pointer to the hci connection request callback
 */
void bt_hw_startup(struct btstack * stack, HCI_CON_RQST_CB);

/*******************************************************************************
 * bt_hw_reset
 ******************************************************************************/
/**
 * @brief Doing a hardware reset of the Bluetooth module
 * 
 * - bt_hw_shutdown()
 * - sleep a while
 * - bt_hw_startup()
 *
 * only available on BTnode3
 *
 * @param stack The bt-stack to use 
 */
void bt_hw_reset(struct btstack * stack);

/*******************************************************************************
 * bt_hci_init
 ******************************************************************************/
/**
 * @brief Initializes hci-layer (specified by given stack argument)
 * 
 * Initializes the bt-stack, dispatches a new thread (the bt-receive thread)
 * and enters (in this thread) an infinite loop to process all the packets.
 *
 * @return pointer of bt stack
 *
 */
struct btstack* bt_hci_init(void *transport);


/*****************************************
 * "Local" Commands
 * -> dont go over the wire!
 * -> Are read directly from mem.
*****************************************/


/*******************************************************************************
 * bt_hci_local_role_discovery
 ******************************************************************************/
/**
 * Returns the Role of the specified conn.handle
 *
 * Does not go over the wire!
 *
 * Do not confuse this command with bt_hci_role_discovery(...)!
 * That cmd. sends a hci-pkt to the host, this (local) cmd
 * just does a lookup in the conn.table!
 *
 * If you opened the connection and allowed a role change, then
 * this information may be inaccurate!
 * After the first call to bt_hci_role_discovery, the local con. is accurate
 * and keeps track of all role changes!
 *
 * #BT_HCI_MY_ROLE_MASTER (this device is master of this connection)
 * #BT_HCI_MY_ROLE_SLAVE (this device is slave of this connection)
 *
 * @param stack The bt-stack to use
 * @param con_handle The handle of the connection
 *
 * @return BT_HCI_MY_ROLE_MASTER or BT_HCI_MY_ROLE_SLAVE or possible errors:
 * - #BT_ERR_NO_CON
 */
long bt_hci_local_role_discovery(struct btstack *stack, bt_hci_con_handle_t con_handle);


/*******************************************************************************
 * bt_hci_local_mode_discovery
 ******************************************************************************/
/**
 * Returns the Mode of the specified conn.handle
 *
 * Does not go over the wire!
 *
 * mode (1 Byte):
 * #BT_HCI_CONN_MODE_ACTIVE
 * #BT_HCI_CONN_MODE_HOLD
 * #BT_HCI_CONN_MODE_SNIFF
 * #BT_HCI_CONN_MODE_PARK
 * 
 * interval (u_short):
 * 0x0000-0xFFFF
 *
 * return (last 3 bytes meaningful):
 * ... | interval | mode |
 * 
 * @param stack The bt-stack to use
 * @param con_handle The handle of the connection
 *
 * @return mode with interval or possible errors:
 * - #BT_ERR_NO_CON
 */
long bt_hci_local_mode_discovery(struct btstack *stack, bt_hci_con_handle_t app_con_handle);

/*******************************************************************************
 * bt_hci_get_remote_cod
 ******************************************************************************/
/**
 * returns the cod of the connection with handle con_handle (24-bit)
 *
 * Does not go over the wire!
 *
 * @param stack The bt-stack to use
 * @param con_handle The handle of the connection
 *
 * @return cod or possible errors:
 * - #BT_ERR_NO_CON
 */
long bt_hci_get_remote_cod(struct btstack *stack, bt_hci_con_handle_t con_handle);


/*******************************************************************************
 * bt_hci_get_con_handle
 ******************************************************************************/
/**
 * once a connection is established the connection will be addressed via a
 * unique connection handle. 
 * The function returns the handle via the argument pointer con_handle (to
 * have a similar API to the inverse function!)
 *
 * Does not go over the wire!
 *
 * @param stack The bt-stack to use
 * @param addr The bdaddr wich is associated with the con_handle
 *
 * @return con_handle or possible errors:
 * - #BT_ERR_NO_CON
 * inverse function: bt_hci_get_remote_bt_addr
 */
long bt_hci_get_con_handle(struct btstack *stack, bt_addr_t addr);

/*******************************************************************************
 * bt_hci_get_local_bt_addr
 ******************************************************************************/
/**
 * The bluetooth address of the local node can be retrieved. 
 *
 * Does not go over the wire!
 *
 * Replaces #bt_hci_read_bt_addr.
 * 
 * @param stack The bt-stack to use
 * @param addr The bt_addr where the result will be stored
 *
 * @return always 0 
 */
long bt_hci_get_local_bt_addr(struct btstack *stack, bt_addr_t addr);

/*******************************************************************************
 * bt_hci_get_remote_bt_addr
 ******************************************************************************/
/**
 * The bluetooth address of the remote node of a connection can be retrieved
 * calling this function with the connection handle. 
 *
 * Does not go over the wire!
 *
 * @param stack The bt-stack to use
 * @param con_handle The handle which is associated with the con_handle
 * @param addr The bt_addr where the result will be stored
 *
 * @return 0 or possible errors:
 * - #BT_ERR_NO_CON
 * inverse function: bt_hci_get_con_handle
 */
long bt_hci_get_remote_bt_addr(struct btstack *stack, bt_hci_con_handle_t con_handle, bt_addr_t addr);


/*******************************************************************************
 * bt_hci_get_con_handles
 ******************************************************************************/
/**
 *
 * Does not go over the wire!
 *
 * This function returns the list of open connections. If only the number
 * of connections is needed, the function can be called with a NULL pointer
 * for the con_handles. Pass NULL instead of con_handles if only interested
 * in number of open connections.
 *
 * @param stack The bt-stack to use
 * @param con_handles Pointer to array with all active connections
 *
 * @return The number of connections (indicating valid position in con_handles)
 */
long bt_hci_get_con_handles(struct btstack *stack, bt_hci_con_handle_array con_handles);


/*******************************************************************************
 * bt_hci_is_con_open
 ******************************************************************************/
/**
 * If the connection with handle con_handle is open, this function returns
 * 0. If the connection is not open anymore, the return value will indicate an error
 * with value BT_ERR_NO_CON.
 *
 * Does not go over the wire!
 *
 * If the connection is in the initialization phase (active or passive),
 * then this function will return BT_ERR_NO_CON, because the connect-process
 * is not jet completed!
 *
 * @param stack The bt-stack to use
 * @param con_handle The handle to check
 *
 * @return 0 or possible errors:
 * - #BT_ERR_NO_CON
 */
long bt_hci_is_con_open(struct btstack *stack, bt_hci_con_handle_t con_handle);


/*******************************************************************************
 * bt_hci_register_con_table_cb
 ******************************************************************************/
/**
 * With this function an application can register a callback function that will
 * be called as soon as the connection table changes. This is the case if:\n
 * -a connection was opened\n
 * -a connection was closed\n
 * -a role change occured\n
 * The callback function will be passed 3 arguments: type of change, details, con_handle:
 * 
 * Type can be:
 * - #BT_HCI_CONNECTION
 * - #BT_HCI_DISCONNECTION
 * - #BT_HCI_MY_ROLE_MASTER 
 * - #BT_HCI_MY_ROLE_SLAVE
 * 
 * Only one callback can be registered at a given time!
 *
 * For #BT_HCI_CONNECTION, details tells whether the connection was opened by us:
 * - #BT_HCI_CONNECTION_ACTIVE     we opened
 * - #BT_HCI_CONNECTION_PASSIVE    other device opened
 * 
 * For #BT_HCI_DISCONNECTION, details give the disconnect reason:
 * - #BT_HCI_DISCON_LINK_LOST
 * - #BT_HCI_DISCON_OTHER_DEVICE          other device requested disconnect
 * - #BT_HCI_DISCON_LOW_RESSOURCE         other device requested and gave reason
 * - #BT_HCI_DISCON_ABOUT_TO_POWER_OFF    other device requested and gave reason
 * - #BT_HCI_DISCON_BY_LOCAL_DEVICE       we requested disconnect
 *
 * 
 *
 * @param stack The bt-stack to use
 * @param #HCI_CON_TABLE_CB The callback that will be registered.
 * @param cb_arg Pointer to a argument.
 */
void bt_hci_register_con_table_cb(struct btstack *stack, HCI_CON_TABLE_CB, void* cb_arg);


/*******************************************************************************
 * bt_hci_send_acl_pkt
 ******************************************************************************/
/**
 * @brief Send a ACL Packet .
 * 
 * Does internal conversion from bt_hci_con_handle_t to baseband tables.\n
 *
 * The HCI-Layer fills the pkt->header fields!\n
 * The higher layer needs to keep track of length (it shall not be larger
 * than stack->acl_max_len or max buffer size)\n
 * This function does NOT fragment the pkts!\n
 *
 * If you want the pkts fragmented, you have to implement it either in this
 * function or handle the pkt->header in a higher layer.\n
 * Most probably the second option would be the better choice!\n
 *
 * If the connection is in the initialization phase (active or passive),
 * then this function will return #BT_ERR_NO_CON, because the connect-process
 * is not yet completed!\n
 *
 * @param stack The bt-stack to use
 * @param con_handle The outgoing handle
 * @param pb_flag Packet Boundary Flag #HCI_PB_FIRST_FRAGMENT or #HCI_PB_CONTD_FRAGMENT
 * @param bc_flag Broadcast Flag #HCI_BC_UNICAST, #HCI_BC_ACTIVE or #HCI_BC_PICONET
 * @param length The length of the payload
 * @param pkt Pointer to the actual pkt, with payload already filled.
 *
 * @return 0 or possible errors:
 * - #BT_ERR_NO_CON
 */
long bt_hci_send_acl_pkt(struct btstack *stack, bt_hci_con_handle_t con_handle, u_char pb_flag, u_char bc_flag, u_short length, struct bt_hci_pkt_acl *pkt);


/*******************************************************************************
 * bt_hci_register_cl_acl_cb
 ******************************************************************************/
/**
 * @brief Registers acl-callback handler for connection-less (channel id = 2) acl packets
 * 
 * HCI and ACL-Layer will exchange pkts. HCI->ACL everytime when a pkt comes in and
 * ACL->HCI everytime when the ACL-layer wants to send a pkt.
 * We dont want the hci layer to run out of free buffers, so the above layer has to
 * hand down a free pkt-buffer everytime an interaction happens.
 * 
 * @param stack The bt-stack to use
 * @param HCI_ACL_CB the functionpointer to the callback
 * @param pkt a first packet that the HCI-layer can hand upwards!
 * @param void pointer that will be passed to the callback function
 */
void bt_hci_register_cl_acl_cb(struct btstack *stack,  hci_acl_cb_t co_acl_cb, struct bt_hci_pkt_acl *pkt, void *arg);

/*******************************************************************************
* bt_hci_register_co_acl_cb
******************************************************************************/
/**
* @brief Registers acl-callback handler for l2cap connection-oriented acl packets
 * 
 * HCI and ACL-Layer will exchange pkts. HCI->ACL everytime when a pkt comes in and
 * ACL->HCI everytime when the ACL-layer wants to send a pkt.
 * We dont want the hci layer to run out of free buffers, so the above layer has to
 * hand down a free pkt-buffer everytime an interaction happens.
 * 
 * @param stack The bt-stack to use
 * @param HCI_ACL_CB the functionpointer to the callback
 * @param pkt a first packet that the HCI-layer can hand upwards!
 * @param void pointer that will be passed to the callback function
 */
void bt_hci_register_co_acl_cb(struct btstack *stack,  hci_acl_cb_t co_acl_cb, struct bt_hci_pkt_acl *pkt, void *arg);


/*******************************************************************************
 * bt_hci_register_sco_cb
 ******************************************************************************/
/**
 * @brief Registers the acl-callback handler
 * 
 * HCI and SCO-Layer will exchange pkts. HCI->SCO everytime when a pkt comes in and
 * SCO->HCI everytime when the SCO-layer wants to send a pkt.
 * We dont want the hci layer to run out of free buffers, so the above layer has to
 * hand down a free pkt-buffer everytime an interaction happens.
 * 
 * @param stack The bt-stack to use
 * @param HCI_SCO_CB the functionpointer to the callback
 * @param pkt a first packet that the HCI-layer can hand upwards!
 */
void bt_hci_register_sco_cb(struct btstack *stack, HCI_SCO_CB, struct bt_hci_pkt_sco *pkt);

/**
 * @brief Registers the accept-connection callback.
 * 
 * Registers a filter function that is called each time a connection request
 * arrives. 
 * 
 * @param stack The bt-stack to use
 * @param HCI_ACCEPT_CON_CB the function pointer to the callback
 * @param cb_arg Pointer to an argument
 * 
 * @see HCI_ACCEPT_CON_CB
 */
void bt_hci_register_accept_con_cb(struct btstack *stack, HCI_CON_RQST_CB, void* cb_arg);

#endif
