/*
 * Copyright (C) 2006 Florian Schuetz <fschuetz@ethz.ch>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * For additional information see http://www.btnode.ethz.ch/
 * 
 * $Id: aodv.h,v 1.1 2007/06/08 09:51:39 bosterm Exp $
 */

/**
 * @file aodv.h
 * @author Florian Schuetz <fschuetz@ethz.ch>
 * @brief Ad-hoc On Demand Vector (AODV) routing implementaion for the
 * CC1000 radio based on [Perkins et al., Ad hoc On-Demand Distance Vector
 * (AODV) Routing]
 *
 * This implementation is designed to work together with the 
 * RoutingForChipcon (RFC) layer. The implementation has been tailored
 * towards the use with btnodes, but could easily be adaptet to Nut/OS.
 * 
 * Much performance can be gained, by adjusting the parameters to the
 * task that the operating set of nodes is assigned to. For a detailed
 * description of the parameters see [Perkins et al., Ad hoc On-Demand 
 * Distance Vector (AODV) Routing] or [Schuetz, Implementation 
 * and Testing of an Ad-hoc Routing Protocol for the Chipcon-Interface 
 * of BTnodes].
 *
 * Further, in debug mode, this implementation alows to change most of
 * the parmameter, inject packages, observe routing states and gather
 * metrics easily. It is therefore well suited to gain insight into
 * routing in distributed, wireless sensor networks.
 *
 * \attention Throughout the code there are coments that say "IMPROVE". 
 * This means that at this point, AODV performance could be improved in
 * a later version.
 */

#ifndef _BT_AODV_H_
#define _BT_ADOV_H_

#include <sys/types.h>
#include <cc/ccc.h>
#include "routing.h"

// ----------------------------------------------------------------------------
// #define constants
// ----------------------------------------------------------------------------

/*
 * AODV Configuration Parameters
 */
#define ACTIVE_ROUTE_TIMEOUT	        10000 ///< How long a unused route is considered active (in ms, paper: 3000ms)
                                              ///<          

#define ALLOWED_HELLO_LOSS		2     ///< Number of HELLO messages that may be lost before route is deactivated.
                                              ///< 

#define BLACKLIST_TIMEOUT		4 * (RREQ_RETRIES * (long)NET_TRAVERSAL_TIME) ///< How long nodes are kept on the 
                                              ///< blacklist. (in ms )
                                              ///< (RREQ_RETRIES * (long)NET_TRAVERSAL_TIME)

#define DELETE_PERIOD			(5 * MAX((long)ACTIVE_ROUTE_TIMEOUT, (long)HELLO_INTERVAL)) ///< Time until an 
                                              ///< inactive route is deleted. (in ms)
                                              ///< 

#define HELLO_INTERVAL			0     ///< Intervall between HELLO messages (if there was no other message). 
                                              ///< Set to 0 to disable HELLO messages. (in ms, paper: 1000ms)
                                              ///< 

#define LOCAL_ADD_TTL			2     ///< Accounts for local delays. (in ms) 
                                              ///< 

#define MAX_REPAIR_TTL			12    ///< Must be 0.3 * NET_DIAMETER. \warning Set by hand to avoid floating 
                                              ///< point calculations. (in ms)
                                              ///< 

#define MY_ROUTE_TIMEOUT		(2 * ACTIVE_ROUTE_TIMEOUT) ///< How long this specific node consideres a route
                                              ///< acitve.
                                              ///< 

//TODO: Adapt this to your network topology
#define NET_DIAMETER			15   ///< Number of nodes on longest path in the network.
                                              ///< 

#define NET_TRAVERSAL_TIME		(2 * NODE_TRAVERSAL_TIME * NET_DIAMETER) ///< How long it takes a message to 
                                              ///< traverse the longest path in the network. (in ms)
                                              ///< 

#define NEXT_HOP_WAIT			(NODE_TRAVERSAL_TIME + 10) ///< How long the processing overhead of the next hop
                                              ///< is expected. (in ms)
                                              ///< 

#define NODE_TRAVERSAL_TIME		200    ///< Time that a packet need to traverse a node. This is calculated by
                                              ///< by the average time a packet needs to be received, processed and sent.
                                              ///< (in ms)
                                              ///< 

#define PATH_DISCOVERY_TIME		(2 * NET_TRAVERSAL_TIME) ///< Worst case thime for discovering a path between two
                                              ///< nodes. (in ms)
                                              ///< 

#define RERR_RATELIMIT			10    ///< Maximum number of RERRs that can be emitted dring an intervall.
                                              ///< 

#define RING_TRAVERSAL_TIME		(2 * NODE_TRAVERSAL_TIME * (TTL_VALUE + TIMEOUT_BUFFER)) ///< Time that is needed
                                              ///< to traverse the ring during expanding ring search. (in ms)
                                              ///< 

#define RREQ_RETRIES			7    ///< Number of route requests sent until destination is accepted unreachable.
                                             ///< 

#define RREQ_RATELIMIT			10   ///< Limits the number of RREQs a node may emit dring an interval.
                                             ///< 

#define TIMEOUT_BUFFER			2    ///< Adds a timeout to the RREP to count for congestion. (in ms)
                                             ///< 

#define	TTL_START			1    ///< The start value for the expanding ring search. (in hops)
                                             ///< 

#define TTL_INCREMENT			2    ///< The increment of the ttl for the expanding ring search. (in hops)
                                             ///< 

#define TTL_THRESHOLD			7    ///< Threshold of the ttl during expanding ring search. (in hops)
                                             ///< 

#define TTL_VALUE			2    ///< ttl field in header during expanding ring search. (in hops)
                                             ///< 

#define ROUTING_TABLE_SIZE		20   ///< Size of the routing table. MUST not be biger than sizeof(char) - 1.
                                             ///< (in hops). 
                                             ///< 
/*
 * Implemenation dependant AODV Configuration Parameters
 */

#define CONTROL_PACKET_QUEUE_SIZE       10   ///< Size of the queues for control packets.
                                             ///< 

#define RREQ_BUFFER_SIZE                50   ///< Size of the buffer for RREQs.
                                             ///< 

#define BLACKLIST_SIZE                  10   ///< Number of nodes that can be blacklisted at the same time.
                                             ///< 

#define DISCOVERY_BUFFER_SIZE           10   ///< Size of the route discovery buffer. (Limits #simultanious discoveries)
                                             ///< 

#define RREP_BUFFER_SIZE                20   ///< Size of the buffer for RREPs that must be ACKed.
                                             ///< 

#define RREP_RETRY                       7   ///< Defines how many times a RREP is resent in absence of RREP-ACKs.
                                             ///< 


// ----------------------------------------------------------------------------
// global variables
// ----------------------------------------------------------------------------
extern u_short aodv_rreq_recv_count;              ///< Number of RREQ packets targeted for this node. (including failed)
extern u_short aodv_rreq_send_count;              ///< Number of RREQ packets sent from this node. (including failed)
extern u_short aodv_rreq_recv_fail_count;         ///< Number of failed RREQ packets passing through this node.
extern u_short aodv_rreq_send_fail_count;         ///< Number of send failures of RREQ packets.
extern u_short aodv_rreq_ignore_count;            ///< Number of ignored RREQs (received twice or more).
extern u_short aodv_rrep_recv_count;              ///< Number of RREP packets targeted for this node.(including failed)
extern u_short aodv_rrep_send_count;              ///< Number of RREP packets sent from this node. (including failed)
extern u_short aodv_rrep_recv_fail_count;         ///< Number of failed RREP packets passing through this node.
extern u_short aodv_rrep_send_fail_count;         ///< Number of send failures of RREP packets.
extern u_short aodv_rrep_ack_recv_count;          ///< Number of RREP-ACK packets targeted for this node.
extern u_short aodv_rrep_ack_send_count;          ///< Number of RREP-ACK packets sent from this node. (including failed)
extern u_short aodv_rrep_ack_send_fail_count;     ///< Number of send failures of RREP-ACK packets.
extern u_short aodv_rerr_recv_count;              ///< Number of RERR packets targeted for this node. 
extern u_short aodv_rerr_send_count;              ///< Number of RERR packets sent from this node. (including failed)
extern u_short aodv_rerr_send_fail_count;         ///< Number of send failures of RERR packets.
extern u_short aodv_invalid_count;                ///< Number of invalid packets targeted for this node.


/**
 * Routing-Interface
 */
extern routing_interface_t aodv_interface;   ///< Interface of the routing protocol.


// ----------------------------------------------------------------------------
// functions
// ----------------------------------------------------------------------------

/**
 * Initialises the AODV library.
 * 
 * @param address [in] mac address of the node.
 * @return 0 on success, -1 otherwise.
 */
int aodv_init(u_short address);

/**
 * Returns the configured mac address.
 * 
 * @return configured mac address.
 */
u_short aodv_get_address(void);

/**
 * Enables LED status reporting.
 * 
 * @param enable_led [in] 0 to disable LED reporting, 1 to enable.
 */ 
void aodv_enable_led(int enable_led);


//
// Terminal Command helpers
//

/**
 * Sets the new node address.
 * 
 * \warning Reboot the node after setting the new address, since behaviour of
 * the MAC protocol on mac change is undefined.
 *
 * @param addr [in] packet length
 * @return -1 if error
 */ 
u_short aodv_set_address(u_short addr);


//
// Debug commands
//
#ifdef ROUTING_DEBUGMODE

/**
 * Sends a custom built RREQ packet
 * 
 * \warning This will most likely mess up routing, if not each parameter
 * reflects the correct state of the network. Only use this for research purposes.
 *
 * @param flags [in] flags of the RREQ acording to Request for Comments 3561
 * @param hopcount [in] number of hops RREQ already made ( = 0 for originator)
 * @param reserved [in] reserved, should be set to 0
 * @param id [in] id of the RREQ. Usually the id of the last RREQ plus one.
 * @param destination_address [in] address of the destination.
 * @param dsn [in] last known sequence number of the destination or 0 and valid flag not set.
 * @param originator_address [in] address of the originator node.
 * @param osn [in] originator sequence number.
 * @param ttl [in] time-to-life in number of hops.
 * @return -1 if error.
 */ 
int aodv_debug_send_rreq_packet(u_char flags, u_char hopcount,
				u_char reserved, u_short id, u_short destination_address,
				u_short dsn, u_short originator_address, u_short osn,
				u_short ttl); 
/**
 * Sends a custom built RREP packet
 * 
 * \warning This will most likely mess up routing, if not each parameter
 * reflects the correct state of the network. Only use this for research purposes.
 *
 * @param addr [in] address where the RREP should be sent.
 * @return -1 if error.
 */ 
int aodv_debug_send_rrep_packet(u_short addr);

/**
 * Sends a custom built RREP-ACK packet
 * 
 * \warning This will most likely mess up routing, if not each parameter
 * reflects the correct state of the network. Only use this for research purposes.
 *
 * @param addr [in] address where the RREP-ACK should be sent.
 * @return -1 if error.
 */ 
int aodv_debug_send_rrep_ack_packet(u_short addr);

/**
 * Sets the AODV sequence number of this node.
 * 
 * \warning This will most likely mess up routing, if the sequence number does not
 * fit in the state of the network. Only use this for research purposes.
 *
 * @param seqno [in] Sequence number that should be used.
 * @return -1 if error.
 */ 
void aodv_set_sequence_no(int seqno);

/**
 * Sets the AODV RREQ id of this node.
 * 
 * \warning This will most likely mess up routing, if the RREQ id does not
 * fit in the state of the network. Only use this for research purposes.
 *
 * @param rreq_id [in] RREQ id that should be used.
 * @return -1 if error.
 */ 
void aodv_set_rreq_id(int rreq_id);

#endif //ROUTING_DEBUGMODE

//
// Testing commands
//

/**
 * Prints the routing table
 * 
 * \attention This command prints the routing table as it would be interpret
 * at the very moment. The actual routing table entries may be different (for
 * example the active flag still set). This is because the table is lazy-evaluated.
 * Since this command should not change the state of the programm, it only displays
 * the results correct, but does not alter routing table entries. Those are altered
 * when the protokoll consults the routing table next times.
 *
 */ 
void aodv_print_rt(void);

/**
 * Prints an entry of the routing table.
 * 
 * \attention This command prints the routing table as it would be interpret
 * at the very moment. The actual routing table entries may be different (for
 * example the active flag still set). This is because the table is lazy-evaluated.
 * Since this command should not change the state of the programm, it only displays
 * the results correct, but does not alter routing table entries. Those are altered
 * when the protokoll consults the routing table next times.
 *
 * @param entry [in] Entry that should be printed.
 */ 
void aodv_print_rt_entry(int entry);

/**
 * Starts a route discovery.
 * 
 * @param dest [in] Destination to which a route shall be built.
 * @return 0 if discovery started successfully.
 */ 
int aodv_route_discover(u_short dst);

/**
 * Get the actual AODV sequence number of this node.
 *
 * @return AODV sequence number.
 */ 
int aodv_get_sequence_no(void);

/**
 * Get the actual RREQ id of this node.
 *
 * @return AODV RREQ id.
 */ 
int aodv_get_rreq_id(void);



/**
 * Routing Interface
 */
extern routing_interface_t aodv_interface;

#endif //_BT_AODV_H_ 
