/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file con_mgr.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief Connection Manager API
 * 
 * Write your application against this interface. Choose an implementation
 * of a connection manager by including the corresponding .h file. E.g. 
 * to use the XTC connection manager, include the file 'cm_xtc.h'.
 */

#ifndef CON_MGR_H
#define CON_MGR_H

#include <stdbool.h>
#include <bt/bt_defs.h>
#include <bt/bt_psm.h>

/**
 * \brief Initializes the Connection Manager
 * 
 * \param bt_stack Pointer to the btstack
 * \param acl_com_stack Pointer to the acl_com_stack
 * \param cm_psm The PSM the connection manager shall use
 */
void con_mgr_init(struct btstack *bt_stack,
					bt_psm_t* psmux,
					u_short cm_psm,
					HCI_CON_TABLE_CB_REGISTRATION,
					u_long cm_cod);

/**
 * \brief Function to register a higher layer callback function.
 * The registered callback function is called when a reliable connection 
 * changes.
 */
void con_mgr_register_con_table_cb(struct btstack* bt_stack, HCI_CON_TABLE_CB, void* cb_arg);

/**
 * \brief Registers terminal commands of the connection manager.
 * Call this function if you need the debug and control commands of the 
 * connection manager.
 */
void con_mgr_register_cmds(void);

/**
 * \brief Disables BT inquiries.
 * Blocks if an inquiry is already in progress.
 */
void con_mgr_inq_disable(void);

/**
 * \brief Reenables BT inquiries.
 */
void con_mgr_inq_enable(void);

/**
 * \brief Sets inquiry period.
 */
void con_mgr_inq_set_period(u_short val);

short con_mgr_fill_con_info(u_char* args, u_char* param);

/**
 * \brief returns an array with reliable connection handles
 */
u_char con_mgr_get_rel_cons(bt_hci_con_handle_t *rel_cons);

/*
 * \brief Returns an info string
 */
char* con_mgr_get_info(void); 

/*
 * \brief Defines if this node is the root node of the topology to be built.
 * Note: this function usually has to be called before calling con_mgr_init().
 */
void con_mgr_set_root(bool mode);

/*
 * \brief Resets the connection manager
 */
void con_mgr_reset(void);

#endif
