/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

 /*
 * btn-led.h
 *
 * 26.05.2004 Martin Hinz <btnode@hinz.ch>
 */

#ifndef LED_H_INCLUDED
#define LED_H_INCLUDED

#include <io.h>
/**

\file btn-led.h

\author Martin Hinz <btnode@hinz.ch>

\date 26.05.2004

\brief Initialization of the on-board LEDs.

The LED's can be turn on and off with the set and clear function. If no LED
is set the heartbeat LED 0 will flash repeatetly to let the user know that
the btnode is still running (power up). This heartbeat can be set differently
with the heartbeat function and can be turned off.\n
There are several patterns that can be displayed (even placed in a queue) with
different speed and repetition to be set.

*/

/*!
 * Max number of pattern that can be stored at once
 */
#define BTN_LED_MAX_PATTERN_QUEUE 5

/*! \name LED Patterns
 * Several patterns can be queued up. During a pattern-flashing all LED
 * activities (setting and clearing) will be stored and shown after the
 * pattern is over.
 */
//@{
/*!
 * No pattern, only heartbeat or other user settings
 */
#define BTN_LED_PATTERN_NON 0

/*!
 * flashing the selected leds once (arg 0-8 for led selection)
 */
#define BTN_LED_PATTERN_ON 1

/*!
 * flashing the selected leds once, after all off for same time (arg 0-8 for led selection)
 */
#define BTN_LED_PATTERN_ON_OFF 2

/*!
 * inverse of #BTN_LED_PATTERN_ON_OFF
 */
#define BTN_LED_PATTERN_OFF_ON 3

/*!
 * smimilar of #BTN_LED_PATTERN_ON_OFF
 */
#define BTN_LED_PATTERN_OFF_ON_OFF 4

/*!
 * each led flashing once going up
 */
#define BTN_LED_PATTERN_UP 5

/*!
 * each led flashing once going down
 */
#define BTN_LED_PATTERN_DOWN 6

/*!
 * knight-rider effect
 */
#define BTN_LED_PATTERN_KNIGHT 7

/*!
 * lower two leds flashing, later upper two leds
 */
#define BTN_LED_PATTERN_HALF 8
//@}
/*!
 * defines infinite time for showing a pattern
 */
#define BTN_LED_INFINITE 0xFF

/*!
 * \brief If 0 then the led pattern queue remains cleared. 
 * Thus only btn_led_set and btn_led_clear is effective.
 */
u_char _btn_led_pattern_enabled;

/*!
 * \brief To clear the pattern queue.
 */
void btn_led_clear_pattern_queue(void);

/*!
\brief  Different patterns can be shown! With different speed and
reiterations. 

This function adds the pattern if a pattern is already showing!\n
Use btn_led_clear_pattern_queue() to delete the pattern queue!\n
Nr can be #BTN_LED_INFINITE (0xFF) to display pattern infinitely till 
btn_led_clear_pattern_queue() is called!
@param pattern One of the pattern described above.
@param arg Some patterns need an argument to select the LEDs
@param speed Speed of the pattern in 100ms (0 not valid)
@param nr Number of repetition of this pattern (can be #BTN_LED_INFINITE)
*/
void btn_led_add_pattern(u_char pattern, u_char arg, u_char speed, u_char nr);


/*!
\brief Hardware setup for LED access.

Call this function to use the on-board LEDs. If thread_on is
set to 1 a new thread called "LED" will be started to display heartbeat and
pattern. Passing 0 to the patterns and heartbeat won't work!
@param thread_on Pass 1 to start the LED-thread to use heartbeat and patterns
*/
void btn_led_init(u_char thread_on);

/*!
\brief  Sets the heartbeat pause and the bit_pattern that will be shown
instead of a heartbeat with given duration.

If this function is not called, a heatbeat with pattern 1 (LED 0) will be
shown, only if no LED is turned on (to indicate that power is still o.k.)
@param pause Quite-Time of heartbeat measured in 100ms
@param bit_pattern 8-bit pattern that will flash as heartbeat
@param duration Flash-duration measured in 100ms
*/
void btn_led_heartbeat(u_char pause, u_char bit_pattern, u_char duration);

/*!
\brief  Turns on LED 0-3
@param nr The LED-Number (0-3) 
*/
void btn_led_set(u_char nr);

/*!
\brief  Turns off LED 0-3
@param nr The LED-Number (0-3) 
*/
void btn_led_clear(u_char nr);

#endif // LED_H_INCLUDED
