#ifndef _SYNC_RTT_H_
#define _SYNC_RTT_H_

 /**
 * @file sync.h
 *  
 * @brief Definition of synchronization functions
 * 
 */
#include <stdio.h>
#include <stdlib.h>				//only for debug printing (we need abs() in _print_dec)

#include <sys/atom.h>
#include <sys/timer.h>

#include <bt/bt_hci_cmds.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_hci_api.h>

#include <led/btn-led.h>

#include <sys/heap.h>

// wrap posix gettimeofday function on unix emulation
#ifdef __BTN_UNIX__
#define gettimeofday(...) BTN_gettimeofday(__VA_ARGS__)
#endif

 /*******************************************************************************
 * FUNCTION LIKE PREPROCESSOR MACROS
 ******************************************************************************/

/**
 * Average of the lower 4 byte of two 64-bit (long long) values, rounded to 32-bit
 */
#define RT_COARSE_AVERAGE(a, b) (((u_long) (a) + (u_long) (b)) >> 1)

 /*******************************************************************************
 * TUNEABLE DCC PARAMETER
 ******************************************************************************/

/** 
 * Maximal drift of clock rate (bounded-drift model) in ppm, scaled by 1'000'000
 */
#define DRIFT_CONSTRAINT 65

/**
 * Adjusts the ratio of maximum to minimum drift: For DRIFT_ADJUST = 0.5, the absolutes of both maximum and minimum drift
 * are equal; the bigger this value is, the faster the upper bound grows.
 * Thus one can take into account knowledge about the average of the clock speed.
 */
#define DRIFT_ADJUST 0.75

/** 
 * Precalculated factor for calculating current bounds
 */
//precalculate factor (1'000'000 * 2^20)/(1'000'000 + DRIFT_CONSTRAINT)
//	in order to replace two long long-divisions by right-shifts
#define MAX_DRIFT_FAC (long int) (1000000.0 / (1000000 - 2 * DRIFT_ADJUST * DRIFT_CONSTRAINT) * 1048576)
/** 
 * Precalculated factor for calculating current bounds
 */
#define MIN_DRIFT_FAC (long int) (1000000.0 / (1000000 + 2 * (1 - DRIFT_ADJUST) * DRIFT_CONSTRAINT) * 1048576)	

/** 
 * Precalculated factor for calculating bounds backwards (delta h < 0)
 */
#define MAX_DRIFT_FAC_BW (long int) (1000000.0 / (1000000 - 2 * (1 - DRIFT_ADJUST) * DRIFT_CONSTRAINT) * 1048576)

/** 
 * Precalculated factor for calculating bounds backwards (delta h < 0)
 */
#define MIN_DRIFT_FAC_BW (long int) (1000000.0 / (1000000 + 2 * DRIFT_ADJUST * DRIFT_CONSTRAINT) * 1048576)	

/**
 * Measured minimum round trip time.
 */
#define SYNC_MIN_DELAY 6

/**
 * Maximum round trip time. Packets arriving with a higher rtt are to be discarded.
 */
#define SYNC_MAX_RTT 120

/** 
 * Length of synchronisation bound in bytes
 */
#define SYNC_BOUND_LEN 4

 /*******************************************************************************
 * ALGORITHM-SPECIFIC PREPROCESSOR CONSTANTS
 ******************************************************************************/
/** 
 * Shortcut for the variable actually containg the currently relevant bounds.
 */
#define SYNC_CURRENT_BOUNDS sync_globals.bounds

#define SYNC_MAX_NEIGHS 10

#define SYNC_DELAY_UNCERTAINTY 10

/** 
 * Maximum computational delay in milliseconds
 */
#define SYNC_COMPUTATIONAL_DELAY 1

/** 
 * Length of the IM timestamp in bytes. Contains full lower and full upper bound.
 * The length of BT_MAX_ACL_COM_PAYLOAD is reduced by this value.
 */	
#define SYNC_TSTAMP_LEN (SYNC_BOUND_LEN + SYNC_BOUND_LEN + 1 + 4)	// length in Bytes!

/** 
 * Marks the beginning of the second bound within a timestamp
 */	
//#define SYNC_TSTAMP_OFFSET_1 6

/** 
 * Marks the beginning of the RTT-timestamp within a timestamp
 */	
#define SYNC_TSTAMP_OFFSET_RTT (SYNC_BOUND_LEN + SYNC_BOUND_LEN + 1)
	
 /*******************************************************************************
 * TYPES
 ******************************************************************************/

/**
 * @struct _sync_nb
 * @brief Represents a neighbour BTnode
 * Make use of the typedef'd type sync_neighbour.\n
 * A linked list is formed out of these elements in order to store the time
 * of the last receiveing/sending event with the corresponding BTnode.
 * The list is always ordered. The BTnode of the latest communication
 * is at the beginning of the list, the BTnode of the oldest communication
 * is at the end. Maximal SYNC_MAX_NEIGHBOURS neighbours are allowed. If 
 * already SYNC_MAX_NEIGHBOURS are stored in the list, the oldest (last)
 * entry will be removed.
*/
typedef struct _sync_nb {
	bt_hci_con_handle_t hdl;
	u_long last_rcvd;
	u_long last_send;
} sync_neigh;

/**
 * @brief Represents a pair of bounds
*/
typedef struct {
	u_long lower;			/**< Lower bound */
	u_long upper;			/**< Upper bound */
	u_long last_update;			/**< Local time when bounds were updated last */
	char validity;				/**< Validity of bounds (0: invalid, 1: valid) */
} syncbounds;

struct _sync_globals_struct {
	struct btstack* bt_stack;
	syncbounds bounds;					///< Global clock bounds
	sync_neigh neighs[BT_HCI_MAX_NUM_CON];
	u_char is_anchor;					///< Switch: anchor node / sensor node
	char bitpattern;					///< Stores the LED pattern
	HCI_CON_TABLE_CB;					///< con table change callback
	void* con_table_cb_arg;				///< con table change callback argument
} sync_globals;

/** 
 * Return type of gettimeofday()
 *
 */
typedef struct _tv timeval;			//should be defined in <time.h> rather than here

/**
 * @struct _tv
 * @brief Represents unix time in seconds and microseconds
*/
struct _tv {
	long tv_sec;         // seconds
	long tv_usec;        // microseconds
};


 /*******************************************************************************
 * GLOBAL VARIABLES
 ******************************************************************************/
#ifdef LLDLY
	u_long lldly_send, lldly_rcvd;
#endif

 /*******************************************************************************
 * FUNCTION PROTOTYPES
 ******************************************************************************/


/*******************************************************************************
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * HIGH LEVEL FUNCTIONS (APPLICATION INTERFACE)
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 ******************************************************************************/

/*******************************************************************************
 * rt_is_valid
 ******************************************************************************/
/**
 * @brief Returns validity of local bounds (High-level function)
 */
char rt_is_valid(void);


/*******************************************************************************
 * rt_get_bounds
 ******************************************************************************/
/**
 * @brief Returns current bounds (High-level function)
 */
syncbounds rt_get_bounds(void);

/*******************************************************************************
 * rt_get_average
 ******************************************************************************/
/**
 * @brief Gives the average of the local bounds (High-level function)
 *
 * Calculates the average of the current local bounds, if validity of bounds is 1.
 * If validity of bounds is 0, the lower bound is passed back.
 *
 * @param av Pointer to the variable where result shall be stored
 */
void rt_get_average(u_long* av);

/*******************************************************************************
 * rt_get_monotonic
 ******************************************************************************/
/**
 * @brief Gives a strictly monotonically increasing clock value. (High-level function)
 *
 * Returns the current lower bound. This value lays within the guarantee interval
 * and is strictly monotonically increasing, though the average error compared
 * to the reference time is maximal.
 *
 * @param millis Pointer to the variable where result shall be stored
 */
void rt_get_monotonic(u_long* millis);

 /*******************************************************************************
 * gettimeofday
 ******************************************************************************/
/**
 * @brief Get current unix time (High-level function)
 * 
 * This is the BTnode implementation of the equally named C-function which
 * gives the current unix time. The time is calculated as the average of the
 * lower and the upper bound. Note that the granularity is in milliseconds, so
 * the last three digits of the microseconds will always be zero. This is because
 * the synchronization is not as accurate as to give microseconds, but the original
 * format was kept for compatibility reasons. 
 *
 * The function will return 1 if the time
 * is valid, 0 if it is invalid, i.e. if the BTnode has valid respective invalide bounds.
 * Communicating with BTnodes having valid bounds will turn the local bounds valid, as there is 
 * piggy-back synchronization.
 *
 * On an anchornode, this function will request for the timestamp from the connected PC.
 *
 * @param *tv Pointer to variable to store the time in.
 * @param *tz Timezone. Not implemented, pass NULL.
 *
 * @return 1 if time is valid, 0 if time is invalid.
 *
 */
int gettimeofday(timeval * tv, void * tz);

/*******************************************************************************
 * rt_get_millis
 ******************************************************************************/
/**
 * @brief Returns current unix milliseconds calculated by averaging the current bounds
 */
u_long rt_get_millis(void);

/*******************************************************************************
 * rt_get_seconds
 ******************************************************************************/
/**
 * @brief Returns current unix seconds calculated by averaging the current bounds
 */
u_long rt_get_seconds(void);

/*******************************************************************************
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * SYNCHRONIZATION FUNCTIONS (INTERNAL)
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
******************************************************************************/


 /*******************************************************************************
 * rt_set_clock (with list initialization)
 ******************************************************************************/
/**
 * @brief Stores bounds into local bounds
 *
 * This function is called when the BTnode receives bounds for the first time. 
 * The own bluetooth address is removed from the linked list and the BTnode which
 * sent the timestamp is inserted as nearest neighbour.
 * The received bounds are stored and their validity is set to 1.\n
 *
 * On an anchornode, this function has no effect as an anchornode doesn't store
 * any bounds.
 */
void rt_set_clock(syncbounds * current_bounds_rcvd, u_long arrival_time, u_short rtt);


 /*******************************************************************************
 * rt_init (IM)
 ******************************************************************************/
/**
 * @brief Initializes the synchronization functions
 *
 * Initializes local bounds to zero. Sets validity of bounds to 1 on anchornodes,
 * to 0 on other nodes. Calls sync_anchor_init() to initialize anchornode.
 */
void rt_init(void);

void rt_set_mode(u_char mode);

/*******************************************************************************
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * COMM-EVENT LAYER FUNCTIONS (BT-HCI FUNCTIONS INTERFACE)
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 ******************************************************************************/

 /*******************************************************************************
 * sync_generate_timestamp
 ******************************************************************************/

/**
 * @brief Called before sending packet
 *
 * This function is called just before a HCI_ACL_DATA_PACKET is sent to the bluetooth
 * module in the function _bt_hci_send_pkt() in bt_hci_transport_uart.h. 
 * It updates the sync_timestamp_to_send in order to transmit the current 
 * timestamp by calling sync_generate_timestamp() with updated bounds.
 *
 * The bluetooth address of the destination device is needed in order to 
 * retrieve the right stored past bounds. If no such bounds exist (communicating for
 * the first time with this BTnode), zero bounds with validity 0 are sent as past
 * bounds.
 *
 * On an anchornode sync_generate_timestamp() is not called as the timestamp is requested
 * from the connected PC and stored in sync_timestamp_to_send directly.
 *
 * @param bt_addr addr The bluetooth address to which the packet is sent to.
 */
void sync_generate_timestamp(bt_hci_con_handle_t hdl, u_char* stamp);


 /*******************************************************************************
 * sync_process_timestamp
 ******************************************************************************/

/**
 * @brief Called after receiving packet
 * 
 * This function is called just after the timestamp of a HCI_ACL_DATA_PACKET 
 * is read from the bluetooth module into sync_rcvd_timestamp in the function
 * _bt_hci_get_pkt() in bt_hci_transport_uart.h.
 * 
 * It calls sync_dissassemble_timestamp() to extract the bounds and intersects
 * them with to localy stored bounds by calling sync_sync_bounds();
 */
void sync_process_timestamp(bt_hci_con_handle_t hdl, u_char* stamp, u_long t_arrive);

void sync_init(struct btstack* bt_stack, HCI_CON_TABLE_CB_REGISTRATION, u_char mode);

void sync_register_con_table_cb(struct btstack* bt_stack, HCI_CON_TABLE_CB, void* cb_arg);

#endif //_SYNC_RTT_H_
