/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * @file bt_time_sync.h
 *
 * @author Matthias Ringwald <mringwal@inf.ethz.ch>
 * @author Qin Yin <qyin@student.ethz.ch>
 *
 * @brief Time Synchronization based on Bluetooth clock offsets
 *
 * This implementation uses l2cap-cl for the one-time exchange of the complete BT clock
 * and provides time-stamped single and multihop send and receive operations
 *
 * To use with mhop, just call bt_time_sync_mhop_init and use bt_time_sync_send_mhop_pkt
 * to send time stamped packets. bt_time_sync_get_time is used to extract time information
 * of a received packets
 *
 * The time sync thread uses the random number generator. We assume that it is already seeded.
 *
 * Details on the implementation can be found in:
 
 * Matthias Ringwald, Kay Rmer: Practical Time Synchronization for Bluetooth Scatternets.
 * Proceedings of the 4th International Conference on Broadband Communications,
 * Networks, and Systems (BROADNETS 2007). 
 * Raleigh, North Carolina, USA, September 2007. 
 * http://www.vs.inf.ethz.ch/publ/papers/mringwal-practi-2007.pdf
 * Invited Paper
 */

#include <sys/types.h>
#include <bt/bt_hci_defs.h>

// PSM for service
#define BT_TIME_CLOCK_PSM 0xFFED

/** basic bt time functions */

/** 
 * low-level init
 *
 * prepares bt time sync service 
 * creates timesync thread which periodically update clock offsets to neighours using inquiry
 * registers connection callback at hci layer
 */
void bt_time_sync_init( struct btstack *bt_stack);

/**
 * process time stamp from neighbour
 * has to be called once per neighbour node
 * this is done by higher-level API automatically (if using bt_time_sync_l2cl_init or bt_time_sync_mhop_init)
 */
void bt_time_sync_handle_neighbor_clock( bt_hci_con_handle_t handle, u_long remote_bt_clock);

/** bt time sync based on l2cap connection-less */

    
/**
 * init bt time sync for use with l2cap connection-less
 *
 * registers data callback for BT_TIME_CLOCK_PSM
 * provides function to send timestamp over l2cap-cl
 *
 * @param bt_stack
 * @param psmux used for l2cap-cl
 */
void bt_time_sync_l2cl_init( struct btstack *bt_stack, bt_psm_t* psmux);

/**
 * send timestamped packet over l2cap-cl
 *
 * @param event_sys_tstamp in milli seconds obtained from NutGetMillis()
 * @param data to send
 * @param len of data
 * @param con handle
 * @param receiver psm
 */
long bt_time_sync_send_pkt(long event_sys_tstamp, u_char *data, u_short data_len, bt_hci_con_handle_t hdl, u_short psm);

/**
 * get time of timestamped event with respect to local time
 *
 * @param pkt received packet
 */
u_long bt_time_sync_get_time( bt_acl_pkt_buf * pkt );

/** bt time sync for mhop routing */

/**
 * init bt time sync for use with mhop 
 * @param bt_stack
 * @param psmux used for l2cap-cl
 * @param mhop_psm the psm of the uses mhop service 
 */
void bt_time_sync_mhop_init( struct btstack *bt_stack, bt_psm_t* psmux, u_short mhop_psm );

/**
 * send timestamped packet over mhop to destination
 * 
 * @param event_sys_tstamp in milli seconds obtained from NutGetMillis()
 * @param data to send
 * @param len of data
 * @param destination. use 
 * @param receiver psm
 * @param bc_flag Determines if the packet should be broadcasted or not. MHOP_CL_BROADCAST | MHOP_CL_UNICAST
 * @param ttl Maximal time to live for the packet - pass TTL_INFINITE for inifinite hops
 */
 long bt_time_sync_send_mhop_pkt(long event_sys_tstamp, u_char* data, u_short data_len, bt_addr_t dest, u_short psm, u_char bc_flag, u_char ttl);
