/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * @file fifo_queue.h
 *
 *\author Kevin Martin <kevmarti@tik.ee.ethz.ch>
 * 
 *\date 11-22-05
 * 
 * \brief Implementation of a first-in first-out (FIFO) queue.
 */

#ifndef _FIFO_QUEUE_H_
#define _FIFO_QUEUE_H_

/**
 * \brief Representation of a FIFO queue.
 * 
 * Contains all the information needed to control a FIFO queue.
 */
typedef struct _fifo_queue_t {
	u_short first;		///< Index of the first element in the queue
	u_short size;		///< Current number of elements the queue contains
	u_short max_size;	///< Pointer to the elements contained in the queue
	void** que_start;	///< Maximal nr of elements the queue is able to handle
} fifo_queue_t;

inline u_short fifo_queue_size(fifo_queue_t* que);

inline u_short fifo_queue_max_size(fifo_queue_t* que);

/**
 * \brief Adds an element to the end of a fifo queue
 * 
 * "Pushes" an element into the fifo queue specified.
 * 
 * \param que Pointer to the fifo queue.
 * \param elem Pointer to the element to be added
 * 
 * \return
 * - 0 if success
 * - -1 if queue is full
 */
inline short fifo_queue_push(fifo_queue_t* que, void* elem);

/**
 * \brief Removes the first element of the FIFO queue specified.
 * 
 * "Pops" the topmost element from the FIFO queue specified.
 * 
 * \param que Pointer to the queue.
 * 
 * \return
 * - pointer to the first element in the queue
 * - NULL if the list is empty
 */
inline void* fifo_queue_pop(fifo_queue_t* que);

/**
 * \brief Gets the first element of the FIFO queue specified.
 * 
 * Returns a pointer to the topmost element in the queue without removing it.
 * 
 * \param que Pointer to the queue.
 * 
 * \return
 * - Pointer to the first element in the queue
 * - NULL if the queue is empty
 */
inline void* fifo_queue_top(fifo_queue_t* que);

fifo_queue_t* fifo_queue_create(u_short size);

void fifo_queue_destroy(fifo_queue_t* que);

#endif //_FIFO_QUEUE_H_
