/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * Time Synchronization based on Bluetooth clock offsets
 * 
 * Matthias Ringwald, Kay Rmer: Practical Time Synchronization for Bluetooth Scatternets.
 * Proceedings of the 4th International Conference on Broadband Communications,
 * Networks, and Systems (BROADNETS 2007). Raleigh, North Carolina, USA, September 2007
 * http://www.vs.inf.ethz.ch/publ/papers/mringwal-practi-2007.pdf
 * 
 * This implementation uses l2cap-cl for the one-time exchange of the complete BT clock
 * and provides time-stamped single and multihop send and receive operations
 *
 * \author Matthias Ringwald <mringwal@inf.ethz.ch>
 * \author Qin Yin <qyin@student.ethz.ch>
 */

/* MR TODO BLOG

1. Problem: Inquiry AND/OR Open Connection seems to have an influence on Nut/BT Clock Conversion:
- which has an influence?
- skip values during page and inquiry

2. Problem: Clock overruns
- nut/os NutGetMillis() counts 1/1024 hz ticks => will overrun after 47 days
- bluetooth clock is extended from 28 bits to 32 bits => will overrung after 15 days

*/

#include <stdio.h>
#include <stdlib.h>     // rand
#include <netinet/in.h> // ntoh etc.
#include <string.h>     // libc memcpy

#include <sys/heap.h>
#include <sys/event.h>
#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/mutex.h>

#include <bt/bt_hci_defs.h>
#include <bt/bt_hci_cmds.h>
#include <bt/bt_acl_defs.h>
#include <bt/bt_psm.h>
#include <bt/l2cap_cl.h>

#include <mhop/mhop_cl.h>
#include <mhop/mhop_cl_defs.h>

// debug settings
#include <debug/syslog.h>
#define LOG_CLASS 0
#define LOG_LEVEL 0
#include <debug/log_set.h>

#include <time/bt_time_sync.h>

/********************************************************/
/* DEFINES */
/********************************************************/

// packet types
#define CLOCK_PKT                   0
#define TSTAMPED_PKT                1
#define MHOP_TSTAMPED_PKT           2

// update rate of bt clock
#define BT_CLOCK_UPDATE_PERIOD_MS 10000 

// median filter size
#define BT_CLOCK_SAMPLES 9

// inquiry period
#define INQ_PERIOD_MS 270000  //  4.5  minutes
#define INQ_RANDOM_MS  65535  // +/- 1 minute
#define INQ_TIME 5            // 5 seconds
#define INQ_MAX_NBS 30        // number of neighbor nodes to expect

/********************************************************/
/* PROTOTYPES                                           */
/********************************************************/

static void bt_time_send_clock ( bt_hci_con_handle_t con_hdl);


/********************************************************/
/* TYPES DEFINITION                                     */
/********************************************************/

typedef struct {
    u_long nut_millis;
    u_long bt_clock;
    u_long bt_millis;
    long   offset;
} bt_clock_sample_t;

typedef struct {    
    bt_hci_con_handle_t hdl;       // handle
    bt_addr_t addr;                // bt address -- remove one of both
    u_long bt_clock_offset;        // bits 2-16 of clock bt_diff 
    long bt_diff;                  // complete difference. add this to get nb clock
    u_char state;                  // state of this connection
    u_long remote_bt_clock;        // received bt clock
} time_sync_neigh_t;

typedef struct {	      
    struct btstack *bt_stack;   // Pointer to lower level hci stack 
    time_sync_neigh_t neighs[BT_HCI_MAX_NUM_CON];  // neighbor table
    
    struct bt_hci_inquiry_result inq_results[INQ_MAX_NBS]; // used during inquiry INQ_MAX_NBS * x bytes..
    HANDLE connection_update;
    bt_clock_sample_t bt_clock_samples[BT_CLOCK_SAMPLES];  // latest samples

    u_long sample_bt_clock;        // reference for conversion bt ticks <-> millis
    u_long sample_nut_millis;      // reference for conversion bt ticks <-> millis

    u_long bt_clock_last;          // used to detect overruns of the bluetooth clock
    u_long bt_clock_high;          // used to correct overruns. 1 overrun = 0x10000000 bt ticks
    u_long bt_millis_high;         // used to correct overruns. 1 overrun = 0x05000000 ms
    
    bt_psm_t* psmux;               // used to deliver received time stamped packets
 
    bt_addr_t addr;                // for mhop routing
    bt_psm_data_cb_t mhop_cb;      // for mhop routing
    u_char * mhop_buffer;          // for mhop routing
    MUTEX    mhop_buffer_mutex;    // for mhop routing
} clock_stack_t;

/** bluetooth clock */
typedef struct {	      
    u_char type;
    long bt_clock;
} clock_pkt_t;

/** single or multi hop data packet with bluetooth clock */
typedef struct {
    u_char type;
    u_short psm;
    long tstamp;          
    
} tstamped_pkt_t;

// Neighbour states
#define NBS_CONNECTED   1
#define NBS_TS_RECEIVED 2
#define NBS_TS_SENT     4
#define NBS_VALID       8



/********************************************************/
/* GLOBAL VARIABLES */
/********************************************************/

clock_stack_t* clock_stack;

/********************************************************/
/* PRIVATE */
/********************************************************/


/**
 * get bt/nut clock sample 
 * Bluetooth clock has only 28 bit but we extend it to 32 bit
 */
static u_long getBtClockSample(void){
    
    // shift samples
    memcpy( &clock_stack->bt_clock_samples[0], &clock_stack->bt_clock_samples[1], sizeof(bt_clock_sample_t) * (BT_CLOCK_SAMPLES-1));
    
    // get sample
    u_long bt_clock = bt_hci_read_clock(clock_stack->bt_stack, BT_HCI_SYNC,0, BT_HCI_CON_HANDLE_OWN, NULL,
                                 &clock_stack->bt_clock_samples[BT_CLOCK_SAMPLES-1].nut_millis);

    // test for overruns as only 28 bits are used
    if ( bt_clock < clock_stack->bt_clock_last) {
        clock_stack->bt_clock_high  += 0x10000000; // 2^28 
        clock_stack->bt_millis_high += 0x05000000; // 2^28 * 5 / 16 = 2^24*5
    }
    clock_stack->bt_clock_last = bt_clock;

    // convert to mills (using bt_millis_high)
    u_long bt_millis = bt_clock * 5 / 16 + clock_stack->bt_millis_high;

    // finally retrieve bits 28-31 (using bt_clock_high)
    bt_clock += clock_stack->bt_clock_high;

    // store new clock sample
    clock_stack->bt_clock_samples[BT_CLOCK_SAMPLES-1].bt_clock  = bt_clock;
    clock_stack->bt_clock_samples[BT_CLOCK_SAMPLES-1].bt_millis = bt_millis;
    clock_stack->bt_clock_samples[BT_CLOCK_SAMPLES-1].offset = bt_millis - clock_stack->bt_clock_samples[BT_CLOCK_SAMPLES-1].nut_millis;
    
    return bt_clock;
}

/**
 * get index of median clock sample
 */
static u_short getMedianSample(void){
    u_char i,j, k;
    u_char sorted[BT_CLOCK_SAMPLES];
    for (i=0;i<BT_CLOCK_SAMPLES;i++) sorted[i] = 1;
    for (j=0; j < (BT_CLOCK_SAMPLES/2+1) ; j++) {
        k = 0xff;
        for (i=0;i<BT_CLOCK_SAMPLES;i++) {
            if (sorted[i] != 0xff) {
                if (k == 0xff || clock_stack->bt_clock_samples[i].offset > clock_stack->bt_clock_samples[k].offset ) {
                    k = i;
                }
            }
        }
        sorted[k] = 0xff;
    }
    return k;
}


/**
 * update bt/nut conversion
 *
 * sample_ .. contains best bt/nut clock sample pair
 */
static void updateBTnutClockConversion(void) {
    
    // get new sample
    getBtClockSample();
    
	u_char best = getMedianSample();
    clock_stack->sample_bt_clock   = clock_stack->bt_clock_samples[best].bt_clock;
    clock_stack->sample_nut_millis = clock_stack->bt_clock_samples[best].nut_millis;
    
    INFO("Update (nut/bt): t2-t1 %ld <= nut millis %lu, bt millis %lu, bt clock %lu\n",
           clock_stack->bt_clock_samples[best].bt_millis-clock_stack->sample_nut_millis,
           clock_stack->sample_nut_millis, clock_stack->bt_clock_samples[best].bt_millis, clock_stack->sample_bt_clock);
}

static time_sync_neigh_t* _time_sync_neigh_add(bt_hci_con_handle_t hdl)
{
    u_char i;
    
    for (i=0; i<BT_HCI_MAX_NUM_CON; i++)
    {
        if (clock_stack->neighs[i].hdl == BT_HCI_HANDLE_INVALID)
        {
            clock_stack->neighs[i].hdl = hdl;
            bt_hci_get_remote_bt_addr (clock_stack->bt_stack, hdl, clock_stack->neighs[i].addr);
            return &clock_stack->neighs[i];
        }
    }
    return NULL;
}

static time_sync_neigh_t* _time_sync_neigh_get_from_hdl(bt_hci_con_handle_t hdl)
{
    u_char i;
	
    for (i=0; i<BT_HCI_MAX_NUM_CON; i++)
    {
        if (clock_stack->neighs[i].hdl == hdl) 
            return &clock_stack->neighs[i];
    }
    return NULL;
}


/**
 * Connection table callback
 * We only handle BT_HCI_CONNECTION and BT_HCI_DISCONNECTION and ignore others.
 * When we get a BT_HCI_CONNECTION event,
 * - add the neighbor to the list of neighbors; 
 * - post the event to the thread. 
 * When we get a BT_HCI_DISCONNECTION event, 
 * - delete the neighbor from the list of neighbors.
 */ 
static void bt_time_con_table_cb(u_char type, u_char reason, bt_hci_con_handle_t con_handle, void* cb_arg)
{  	
    time_sync_neigh_t * nb;
    if (type == BT_HCI_CONNECTION) 
    {
        INFO("INFO: Connection opened (%d)\n",con_handle);  

        // enter into neighbour table
        nb = _time_sync_neigh_add( con_handle );
        
        // set state
        nb->state = NBS_CONNECTED;
        
        // Signal time synchronization thread
        NutEventPost(&clock_stack->connection_update);
    } 
    else if (type == BT_HCI_DISCONNECTION) 
    {
        INFO("INFO: Connection closed (%d)\n",con_handle);     	

        // remove from neighbour table
        nb = _time_sync_neigh_get_from_hdl( con_handle );
        nb->hdl = BT_HCI_HANDLE_INVALID;
        nb->state = 0;
    } 
}




/**
 * process received timestamp for neighbor
 * - update clock offset
 * - transform to local clock
 * - forward updated time stamp
 *
 * calculate clock offset to neighbour node
 * input:
 * local  bt clock
 * clock information for neighbor (remote bt clock, role, clock offset)
 *
 * output:
 * clock offset to remote node
 */
static void processTimestamp(time_sync_neigh_t *nb, u_long local_bt_clock){
    
    long res = bt_hci_read_clock_offset(clock_stack->bt_stack, BT_HCI_SYNC, nb->hdl);
    nb->bt_clock_offset = res << 2;
    INFO("bt_hci_read_clock_offset: %ld\n", res);

    long offsetToRemote;

    // calc clock difference
    u_char role = (u_char) bt_hci_local_role_discovery( clock_stack->bt_stack, nb->hdl);
    u_long clock_offset_approx = 0;
    if (role == BT_HCI_MY_ROLE_SLAVE) {
        clock_offset_approx = local_bt_clock - nb->remote_bt_clock;
        if ( (clock_offset_approx & 0x1ffff) < nb->bt_clock_offset){
            clock_offset_approx -= 0x20000;
        }
        long clock_offset_slaveToMaster = (clock_offset_approx & 0xfffe0000) | nb->bt_clock_offset;
        offsetToRemote = - clock_offset_slaveToMaster;
    } else {
        clock_offset_approx = nb->remote_bt_clock - local_bt_clock;
        if ( (clock_offset_approx & 0x1ffff) > nb->bt_clock_offset){
            clock_offset_approx += 0x20000;
        }
        long clock_offset_slaveToMaster = (clock_offset_approx & 0xfffe0000) | nb->bt_clock_offset;
        offsetToRemote = clock_offset_slaveToMaster;
    }
    INFO("Remote Clock: %ld, Local Clock %ld, diff %ld \n", nb->remote_bt_clock, local_bt_clock, offsetToRemote);
    
    // finally
    nb->bt_diff = offsetToRemote;
}


/**
 * update clock offsets based on previous clock differences
 * 
 * assumption clock drifts rather than jumps
 * pre: nb->state |= NBS_VALID (high bits are valid)
 *
 * could be moved into general HCI handler. store Clock Offset for connected nbs
 * complete inq_results table not necessary anymore
 * then, we would not know, when to update. but computation is cheap anyway.
 *
 */
static void updateClockOffsets(void){

    long res;
    u_char nr_results;
    u_char i;
    time_sync_neigh_t* nb;
    u_long clock_offset;
    long new_clock_diff;
    
    INFO("updateClockOffsets: ");
    
    // inquiry for INQ_TIME * 1.28 s
    res = bt_hci_inquiry( clock_stack->bt_stack, BT_HCI_SYNC, INQ_TIME, INQ_MAX_NBS, &clock_stack->inq_results[0]);
    INFO("res %ld neigbhours\n", res);
    if (res < 0) {
        return;
    }
    nr_results = (u_char) res;
    // update neighbor delta
    for (i=0; i < nr_results; i++) {
        res = bt_hci_get_con_handle( clock_stack->bt_stack, clock_stack->inq_results[i].bdaddr);
        clock_offset = clock_stack->inq_results[i].clock_offset & 0x7fff;
        if (res != BT_ERR_NO_CON) {
            nb = _time_sync_neigh_get_from_hdl( (bt_hci_con_handle_t) res );
            if (nb && (nb->state & NBS_VALID)) {
                INFO("inq: handle %u offset %lu \n",  nb->hdl, clock_offset);
                // no need to check for role: clock offsets on inquiry is independent from connection 
                // use 0.3125 us units
                clock_offset <<= 2;
                new_clock_diff = (nb->bt_diff & 0xfffe0000) | clock_offset;
                // correct over/underruns
                if (new_clock_diff >= nb->bt_diff + 0x10000) {
                    new_clock_diff -= 0x20000;
                }
                if (new_clock_diff <= nb->bt_diff - 0x10000) {
                    new_clock_diff += 0x20000;
                }
                nb->bt_diff = new_clock_diff;
                INFO("New bt_diff of %ld to %d\n", new_clock_diff, (u_short) res); 
            }
        }
    }
}


/** 
 * Periodic time services
 *
 * - update nut/bt clock conversion (only when connected?)
 * - send bt clock stamps to new neighbours
 * - get clock offset of new neighbours
 * - inquiries to update clock offset to neigbhours
 */
THREAD( BT_TIME_SYNC, arg ) {
    u_long nextClockUpdate = 0;
    u_long nextInquiry = 0;
    u_long nextAction;
    u_char paused = 0;
    u_long time;
    u_long local_bt_clock;
    u_char i;
    
    while (1) {

        time = NutGetMillis();
        local_bt_clock = 0;
        
        // handle received time stamps first ( CONNECTED, RECEIVED, !VALID)
        for (i=0; i<BT_HCI_MAX_NUM_CON; i++)
        {
            if ( (clock_stack->neighs[i].state & ( NBS_CONNECTED | NBS_TS_RECEIVED | NBS_VALID )) == (NBS_CONNECTED | NBS_TS_RECEIVED) ) {
                if (local_bt_clock == 0) {
                    local_bt_clock = getBtClockSample();
                }
                processTimestamp( &clock_stack->neighs[i], local_bt_clock);
                clock_stack->neighs[i].state |=  NBS_VALID;
            }
        }
        
        // then sent time stamps (CONNECTED, !SENT)
        for (i=0; i<BT_HCI_MAX_NUM_CON; i++)
        {
            if ( (clock_stack->neighs[i].state & (NBS_CONNECTED | NBS_TS_SENT)) == NBS_CONNECTED ) {
                
                // pause before sending over newly opened connections
                // TODO: might be fixed at lower level. packet for non-existing handle => connection was opened
                if (paused == 0) {
                    NutSleep ( 1000 );
                    paused = 1;
                }
                
                // send bt clock
                bt_time_send_clock(clock_stack->neighs[i].hdl);
                
                // update state
                clock_stack->neighs[i].state |=  NBS_TS_SENT;
            }
        }
        
        // Update BT clock
        if (time >= nextClockUpdate) {
            updateBTnutClockConversion();
            nextClockUpdate = time + BT_CLOCK_UPDATE_PERIOD_MS;
        }

        // clock offset update by inquiry
        if (time >= nextInquiry) {
            updateClockOffsets();
            nextInquiry = time + INQ_PERIOD_MS + (rand() & INQ_RANDOM_MS);
        }

        // sleep until next action
        time = NutGetMillis();
        if (nextInquiry < nextClockUpdate) {
            nextAction = nextInquiry;
        } else {
            nextAction = nextClockUpdate;
        }
        if (nextAction > time) {
            NutEventWait( &clock_stack->connection_update, nextAction - time);
        }
    }
}

//////////////////////////////////////////////////////////////////////
// 
//  Low-Level Public API
// 
//

/**
 *
 *
 */
void bt_time_sync_init( struct btstack *bt_stack) {
    u_char i;
    
    // register connection handler callbacks
    bt_hci_register_con_table_cb(bt_stack, &bt_time_con_table_cb, NULL);
    
    // alloc clock service data
    clock_stack = NutHeapAllocClear(sizeof(clock_stack_t));	   
    clock_stack->bt_stack = bt_stack;
    
    // init neigbhour table
    for (i=0; i<BT_HCI_MAX_NUM_CON; i++)
    {
        clock_stack->neighs[i].hdl = BT_HCI_HANDLE_INVALID;
        clock_stack->neighs[i].state = 0;
    }
    
    // get initial clock/nut samples
    for (i=0;i<BT_CLOCK_SAMPLES;i++){
        getBtClockSample();
    }
    
    // start thread
    NutThreadCreate("TIMESYNC", BT_TIME_SYNC, 0, 512);
}

/**
 * store bluetooth clock of neighbor
 *
 * provides a transport indenpentent way to set this value
 */
void bt_time_sync_handle_neighbor_clock( bt_hci_con_handle_t handle, u_long remote_bt_clock){
	// Make sure we can find the neighbor, so that the event timestamps can be transfered. 
    time_sync_neigh_t * nb = _time_sync_neigh_get_from_hdl(handle);
    if (!nb)
    {
        ERROR("ERROR: can't find neighbor in _time_sync_pkt_cb!\n");
        return;
    }
    // store nb clock
    
    nb->remote_bt_clock = remote_bt_clock;
    // update state
    nb->state |= NBS_TS_RECEIVED;
    
    // Signal time synchronization thread
    NutEventPost(&clock_stack->connection_update);
}

/**
* convert a local nut_millis value to bluetooth clock ticks
 * 
 * 1 bt clock tick  = 5/16 millis (0.3125 ms)
 */
u_long bt_time_sync_get_bt_clock_for ( u_long nut_millis ){
    long deltaNutMillis = nut_millis - clock_stack->sample_nut_millis;
    long deltaBtTicks = (deltaNutMillis * 16) / 5;
    u_long result = clock_stack->sample_bt_clock + deltaBtTicks;
    return result;
}

/**
* convert a bluetooth time to nut millis
 * 
 * 1 bt clock tick  = 5/16 millis (0.3125 ms)
 */
u_long bt_time_sync_get_sys_clock_for ( u_long bt_time ){
    long deltaBtTicks = bt_time - clock_stack->sample_bt_clock;
    long deltaNutMillis = (deltaBtTicks * 5) / 16;
    u_long result = clock_stack->sample_nut_millis + deltaNutMillis;
    return result;
}

/**
 * get offset to neighbor
 *
 * @param hdl handle to neighbor
 * @return difference to neighbor in BT ticks (0.3125 ms)
 *
 */
long bt_time_sync_get_offset_to( bt_hci_con_handle_t hdl) {
    time_sync_neigh_t *nb = _time_sync_neigh_get_from_hdl( hdl );
    return nb->bt_diff;
}


//////////////////////////////////////////////////////////////////////
// L2CAP CL

// PRIVATE


/**
 * \brief Data callback of the "time sync" service.
 * 
 * This callback is called CLOCK_PKT and single hop
 * time stamped packets
 * 
 * \param msg Pointer to the received packet
 * \param data Pointer to the payload of the received packet
 * \param data_len Size of the payload
 * \param service_nr The service number this callback corresponds to
 * \param cb_arg Callback argument: Unused in this example.
 */
static bt_acl_pkt_buf* _time_sync_pkt_cb(bt_acl_pkt_buf* pkt_buf,
                                         u_char* data,
                                         u_short data_len,
                                         u_short service_nr,
                                         void* cb_arg)
{	
    u_long remote_bt_clock;
    tstamped_pkt_t * header;
	// Get neibghbour handle
    bt_hci_con_handle_t hdl = bt_acl_get_con_handle(pkt_buf->pkt);
    
    switch ( data[0] ) {
        
        case CLOCK_PKT:
            
            remote_bt_clock = ntohl( *(u_long *) &data[1]);
            bt_time_sync_handle_neighbor_clock( hdl, remote_bt_clock );
            INFO("Hdl %u, remote clock %lu", hdl, remote_bt_clock);
            break;
            
        case TSTAMPED_PKT:
        case MHOP_TSTAMPED_PKT:
            
            header = (tstamped_pkt_t*) data;
            u_long local_bt_clock = ntohl( header->tstamp );
            if (data[0] == TSTAMPED_PKT) { 
                // get neighbor info
                time_sync_neigh_t * nb  = _time_sync_neigh_get_from_hdl( hdl );
                if (nb == 0 || (nb->state & NBS_VALID) == 0) {
                    ERROR("Error: no time diff for handle (%u)!\n", hdl);
                    return pkt_buf;
                }
                // transform remote bt time into local bt time
                // we do the clock transformation on the receiver side to be able to handle broadcasts, too.
                local_bt_clock -= nb->bt_diff;
                header->tstamp = htonl( local_bt_clock) ;
            }
            u_short psm = ntohs(header->psm);
            INFO("TSTAMPED DATA: buf %x, data %x, for psm %u, local bt time %lu\n", (u_short) pkt_buf, (u_short) data, psm, local_bt_clock);
            
            // remove tstamp header
            data     += sizeof(tstamped_pkt_t);
            data_len -= sizeof(tstamped_pkt_t);
            // update buffer, too.
            pkt_buf->pload = data;
            pkt_buf->plen  = data_len;
            
            // else, deliver the message
            bt_acl_pkt_buf* next_buf;
            long retval = bt_psm_deliver_msg(clock_stack->psmux, pkt_buf, data, data_len, psm, &next_buf);
            if (retval)
            {
                ERROR("Error: delivering pkt to psm %d: %ld!\n", psm, retval);				
            } else {
                return next_buf;
            }
            break;

            
    }
    return pkt_buf;
}

/**
 * get current BT clock and send over l2cap cl
 */
static void bt_time_send_clock ( bt_hci_con_handle_t con_hdl) {
    u_long bt_clock = getBtClockSample();
    u_char packet[5];
    packet[0] = CLOCK_PKT;
    *(u_long *) &packet[1] = htonl(bt_clock);
    l2cap_cl_send(packet, sizeof(packet), con_hdl, BT_TIME_CLOCK_PSM);
}

// PUBLIC

/**
 * get timestamp of sent packet
 *
 * assumption: the time sync header exists before the actual data
 *
 * @param pkt_buf of tstamp or mhop_tstamp packet
 * @return time in local nut millis
 */
u_long bt_time_sync_get_time(bt_acl_pkt_buf* pkt ){
    u_char * ptr = pkt->pload - sizeof( long ); // quite ugly, actually.
    long bt_time = ntohl( *(u_long*) ptr);
    return bt_time_sync_get_sys_clock_for( bt_time );
}



/**
* \brief Sends a timestamped event data packet.
 * 
 * \param event_bt_tstamp The local time stamp of the event.
 * \param data Pointer to the data to send
 * \param data_len Amount of data to send
 * \param hdl Connection handle the data shall be sent to 
 * \param psm Identifies the service the data shall be delivered to on the
 * receiving side
 * 
 * \return 0 or possible errors.
 */

/**
 * helper to send a packet (= header + data) to a specific psm on a handle
 */
static inline long _send_header_and_packet( u_char *header, u_char header_len, u_char *data,
                                     u_short data_len, bt_hci_con_handle_t hdl, u_short psm) {
    
    u_short packet_len = header_len + data_len;
    
    // check len
    if (packet_len > l2cap_cl_max_payload() )
    {
        ERROR("Error: not enough space to add the timestamp!\n");
        return BT_ERR_ACL_PLOAD_OVFLW;
    }
    
    // acquire l2cap cl channel
    long retval = 0; 
    retval = l2cap_cl_acquire_write_lock(hdl, psm, packet_len);    	
    if (retval)
    {
        ERROR("Error: open l2cap cl chan (%ld)!\n", retval);
        return retval;
    }    
	
    // write header     
    retval = l2cap_cl_lowlevel_write( header, header_len );
    if (retval)
    {
        ERROR("Error: sending header: %ld\n", retval);
        l2cap_cl_release_write_lock();
        return retval;
    }
    
    // write data
    retval = l2cap_cl_lowlevel_write(data, data_len);
    if (retval)
        ERROR("Error: send data (%ld)!\n", retval);		
    
    // close channel
    l2cap_cl_release_write_lock();    
    
    return retval;
}

long bt_time_sync_send_pkt(long event_sys_tstamp, 
                       u_char *data,
                       u_short data_len,				     		      
                       bt_hci_con_handle_t handle,  
                       u_short psm) 
{ 
    u_long local_bt_time =  bt_time_sync_get_bt_clock_for( event_sys_tstamp );
    INFO("bt_time_sync_send: event %lu, local bt %lu\n",  event_sys_tstamp, local_bt_time);
    
    // build header
    tstamped_pkt_t header;
    header.type   = TSTAMPED_PKT;
    header.psm    = htons(psm);
    header.tstamp = htonl(local_bt_time);    	
    
    return _send_header_and_packet( (u_char*) &header, sizeof(tstamped_pkt_t), data, data_len, handle, psm);
}

/**
 * init bluetooth timesync using mhop over l2cap connection-less data
 */
void bt_time_sync_l2cl_init( struct btstack *bt_stack, bt_psm_t* psmux ) {
    
    // init lower layer time sync
    bt_time_sync_init( bt_stack);
    
    // store psmux 
    clock_stack->psmux = psmux;
    
    // store local address
    bt_hci_read_bt_addr(clock_stack->bt_stack, BT_HCI_SYNC, clock_stack->addr);

    // register our psm  with bt_psm using fixed psm id
    bt_psm_service_register(psmux, BT_TIME_CLOCK_PSM, _time_sync_pkt_cb, NULL);
}



//////////////////////////////////////////////////////////////////////
// MHOP


// PRIVATE

/**
 * \brief Data callback of the "time sync" service.
 * 
 * This callback is called before for each incoming mhop packet
 * 
 * \param msg Pointer to the received packet
 * \param data Pointer to the payload of the received packet
 * \param data_len Size of the payload
 * \param service_nr The service number this callback corresponds to
 * \param cb_arg Callback argument: Unused in this example.
 */
static bt_acl_pkt_buf* _time_sync_mhop_pkt_cb(bt_acl_pkt_buf* pkt_buf,
                                                 u_char* data,
                                                 u_short data_len,
                                                 u_short service_nr,
                                                 void* cb_arg)
{	

    // get pointer to multi-hop packet header
    mhop_cl_pkt_hdr * header = (mhop_cl_pkt_hdr*) data;
    // get psm the packet has to be delivered to
    u_short psm = header->psm[0] | header->psm[1] << 8;

    bt_hci_con_handle_t in_hdl = bt_acl_get_con_handle(pkt_buf->pkt);
    INFO("MHOP INTERCEPTED. inhdl = %u, handle own = %u, equal %d\n", in_hdl, BT_HCI_CON_HANDLE_OWN, (in_hdl == BT_HCI_CON_HANDLE_OWN ));

    // check MHOP PSM
    if (psm == BT_TIME_CLOCK_PSM) {

        // Get neighbour handle
        if (in_hdl != BT_HCI_CON_HANDLE_OWN) {
            
            // get bt diff to nb
            time_sync_neigh_t *nb  = _time_sync_neigh_get_from_hdl( in_hdl );
            if (nb == 0 || (nb->state & NBS_VALID) == 0) {
                ERROR("Error: no time diff for handle (%u)!\n", in_hdl);
                return pkt_buf;
            }
            // get access to tstamped_pkt_t after mhop header
            u_char * payload = data + sizeof(mhop_cl_pkt_hdr);
            tstamped_pkt_t * mh_header = (tstamped_pkt_t*) payload;
            // update tstamp for local bt time
            // we do the clock transformation on the receiver side to be able to handle broadcasts, too
            u_long local_bt_clock = ntohl( mh_header->tstamp ) - nb->bt_diff;
            mh_header->tstamp = htonl( local_bt_clock);
            INFO("MHOP DATA: local bt time %lu\n", local_bt_clock);
        }
    }
    
    // execute original mhop handler
    return clock_stack->mhop_cb( pkt_buf, data, data_len, service_nr, cb_arg);
}


// PUBLIC

/**
 * \brief Sends a timestamped event data packet through mhop.
 *
 * \param event_sys_tstamp The nut time stamp of the event.
 * \param data Pointer to the data to send
 * \param len Amount of data to send
 * \param dest BT address of the destination
 * \param psm Identifies the service the data shall be delivered to on the
 * receiving side
 * \param bc_flag Determines if the packet should be broadcasted or not. 
 * MHOP_CL_BROADCAST | MHOP_CL_UNICAST
 * \param ttl Maximal time to live for the packet - pass TTL_INFINITE for inifinite hops
 * 
 * \return 0 or possible errors.
 */
long bt_time_sync_send_mhop_pkt(long event_sys_tstamp,
                            u_char* data,
                            u_short data_len,
                            bt_addr_t dest,
                            u_short psm,
                            u_char bc_flag,
                            u_char ttl)
{
    // check len
    u_short len = data_len + sizeof(tstamped_pkt_t);
    if (len > l2cap_cl_max_payload() ) {
        ERROR("Error: not enough space to add the timestamp!\n");
        return BT_ERR_ACL_PLOAD_OVFLW;
    }
    
    // lock buffer to concat mhop_tstamped_pkt_t header and data
    NutMutexLock( &clock_stack->mhop_buffer_mutex );

    // create tstamped header
    tstamped_pkt_t * header = (tstamped_pkt_t*) clock_stack->mhop_buffer;
    u_long local_bt_time =  bt_time_sync_get_bt_clock_for( event_sys_tstamp );
    INFO("bt_time_sync_send: event %lu, local bt %lu\n",  event_sys_tstamp, local_bt_time);
    header->type   = MHOP_TSTAMPED_PKT;
    header->psm    = htons(psm);
    header->tstamp = htonl(local_bt_time);    	

    // append data
    memcpy ( &clock_stack->mhop_buffer[sizeof(tstamped_pkt_t)], data, data_len);
    
    // send packet
    long retval = mhop_cl_send_pkt ( clock_stack->mhop_buffer, len, dest, BT_TIME_CLOCK_PSM, bc_flag, ttl);
    
     // lock buffer to concat mhop_tstamped_pkt_t header and data
     NutMutexUnlock( &clock_stack->mhop_buffer_mutex );

     return retval;
}


/**
 * init bluetooth timesync using mhop over l2cap connection-less data
 */
void bt_time_sync_mhop_init( struct btstack *bt_stack, bt_psm_t* psmux, u_short mhop_psm ) {
    
    // init lower layer time sync
    bt_time_sync_l2cl_init( bt_stack, psmux);
    
    // alloc mhop buffer
    clock_stack->mhop_buffer = NutHeapAllocClear(l2cap_cl_max_payload());
    NutMutexInit( &clock_stack->mhop_buffer_mutex );
    
    // intercept mhop registration by our service
    clock_stack->mhop_cb = bt_psm_service_patch( psmux, mhop_psm, _time_sync_mhop_pkt_cb);
}
